# Authors: Karl MacMillan <kmacmillan@mentalrootkit.com>
#          Jan Cholasta <jcholast@redhat.com>
#
# Copyright (C) 2007-2013  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import print_function, absolute_import

import os
import os.path
import tempfile
import optparse  # pylint: disable=deprecated-module

from ipalib import x509
from ipalib.install import certmonger
from ipaplatform.paths import paths
from ipapython import admintool
from ipapython.certdb import NSSDatabase, get_ca_nickname
from ipapython.dn import DN
from ipapython import ipaldap
from ipalib import api, errors
from ipaserver.install import certs, dsinstance, installutils, krbinstance


class ServerCertInstall(admintool.AdminTool):
    command_name = 'ipa-server-certinstall'

    usage = "%prog <-d|-w|-k> [options] <file> ..."

    description = "Install new SSL server certificates."

    @classmethod
    def add_options(cls, parser):
        super(ServerCertInstall, cls).add_options(parser)

        parser.add_option(
            "-d", "--dirsrv",
            dest="dirsrv", action="store_true", default=False,
            help="install certificate for the directory server")
        parser.add_option(
            "-w", "--http",
            dest="http", action="store_true", default=False,
            help="install certificate for the http server")
        parser.add_option(
            "-k", "--kdc",
            dest="kdc", action="store_true", default=False,
            help="install PKINIT certificate for the KDC")
        parser.add_option(
            "--pin",
            dest="pin", metavar="PIN", sensitive=True,
            help="The password of the PKCS#12 file")
        parser.add_option(
            "--dirsrv_pin", "--http_pin",
            dest="pin",
            help=optparse.SUPPRESS_HELP)
        parser.add_option(
            "--cert-name",
            dest="cert_name", metavar="NAME",
            help="Name of the certificate to install")
        parser.add_option(
            "-p", "--dirman-password",
            dest="dirman_password",
            help="Directory Manager password")

    def validate_options(self):
        super(ServerCertInstall, self).validate_options(needs_root=True)

        installutils.check_server_configuration()

        if not any((self.options.dirsrv, self.options.http, self.options.kdc)):
            self.option_parser.error(
                "you must specify dirsrv, http and/or kdc")

        if not self.args:
            self.option_parser.error("you must provide certificate filename")

    def ask_for_options(self):
        super(ServerCertInstall, self).ask_for_options()

        if not self.options.dirman_password:
            self.options.dirman_password = installutils.read_password(
                "Directory Manager", confirm=False, validate=False, retry=False)
            if self.options.dirman_password is None:
                raise admintool.ScriptError(
                    "Directory Manager password required")

        if self.options.pin is None:
            self.options.pin = installutils.read_password(
                "Enter private key unlock",
                confirm=False, validate=False, retry=False)
            if self.options.pin is None:
                raise admintool.ScriptError(
                    "Private key unlock password required")

    def run(self):
        api.bootstrap(in_server=True, confdir=paths.ETC_IPA)
        api.finalize()
        api.Backend.ldap2.connect(bind_pw=self.options.dirman_password)

        if self.options.dirsrv:
            self.install_dirsrv_cert()

        if self.options.http:
            self.replace_http_cert()

        if self.options.kdc:
            self.replace_kdc_cert()

        print(
            "Please restart ipa services after installing certificate "
            "(ipactl restart)")

        api.Backend.ldap2.disconnect()

    def install_dirsrv_cert(self):
        serverid = ipaldap.realm_to_serverid(api.env.realm)
        dirname = dsinstance.config_dirname(serverid)

        conn = api.Backend.ldap2
        entry = conn.get_entry(DN(('cn', 'RSA'), ('cn', 'encryption'),
                                  ('cn', 'config')),
                               ['nssslpersonalityssl'])
        old_cert = entry.single_value['nssslpersonalityssl']

        server_cert = self.import_cert(dirname, self.options.pin,
                                       old_cert, 'ldap/%s' % api.env.host,
                                       'restart_dirsrv %s' % serverid)

        entry['nssslpersonalityssl'] = [server_cert]
        try:
            conn.update_entry(entry)
        except errors.EmptyModlist:
            pass

    def replace_http_cert(self):
        """
        Replace the current HTTP cert-key pair with another one
        from a PKCS#12 file
        """
        # pass in `host_name` to perform
        # `NSSDatabase.verify_server_cert_validity()``
        cert, key, ca_cert = self.load_pkcs12(
            ca_chain_fname=paths.IPA_CA_CRT,
            host_name=api.env.host
        )

        key_passwd_path = paths.HTTPD_PASSWD_FILE_FMT.format(host=api.env.host)

        req_id = self.replace_key_cert_files(
            cert, key,
            cert_fname=paths.HTTPD_CERT_FILE,
            key_fname=paths.HTTPD_KEY_FILE,
            ca_cert=ca_cert,
            passwd_fname=key_passwd_path,
            cmgr_post_command='restart_httpd')

        if req_id is not None:
            certmonger.add_principal(
                req_id, 'HTTP/{host}'.format(host=api.env.host))
            certmonger.add_subject(req_id, cert.subject)

    def replace_kdc_cert(self):
        # pass in `realm` to perform `NSSDatabase.verify_kdc_cert_validity()`
        cert, key, ca_cert = self.load_pkcs12(
            ca_chain_fname=paths.CA_BUNDLE_PEM, realm_name=api.env.realm)

        self.replace_key_cert_files(
            cert, key, paths.KDC_CERT, paths.KDC_KEY, ca_cert,
            profile="KDCs_PKINIT_Certs"
        )

        krb = krbinstance.KrbInstance()
        krb.init_info(
            realm_name=api.env.realm,
            host_name=api.env.host,
        )
        krb.pkinit_enable()

    def load_pkcs12(self, ca_chain_fname=paths.IPA_CA_CRT, **kwargs):
        # Note that the "installutils.load_pkcs12" is quite a complex function
        # which performs some checking based on its kwargs:
        #       host_name performs NSSDatabase.verify_server_cert_validity()
        #       realm performs NSSDatabase.verify_kdc_cert_validity()
        pkcs12_file, pin, ca_cert = installutils.load_pkcs12(
            cert_files=self.args,
            key_password=self.options.pin,
            key_nickname=self.options.cert_name,
            ca_cert_files=[ca_chain_fname],
            **kwargs)

        # Check that the ca_cert is known and trusted
        with tempfile.NamedTemporaryFile() as temp:
            certs.install_pem_from_p12(pkcs12_file.name, pin, temp.name)
            cert = x509.load_certificate_from_file(temp.name)

        with tempfile.NamedTemporaryFile("rb") as temp:
            certs.install_key_from_p12(pkcs12_file.name, pin, temp.name)
            key = x509.load_pem_private_key(
                temp.read(), None, backend=x509.default_backend())

        return cert, key, ca_cert

    def replace_key_cert_files(
        self, cert, key, cert_fname, key_fname, ca_cert, passwd_fname=None,
        profile=None, cmgr_post_command=None
    ):
        try:
            ca_enabled = api.Command.ca_is_enabled()['result']
            if ca_enabled:
                certmonger.stop_tracking(certfile=cert_fname)

            pkey_passwd = None
            if passwd_fname is not None:
                with open(passwd_fname, 'rb') as f:
                    pkey_passwd = f.read()

            x509.write_certificate(cert, cert_fname)
            x509.write_pem_private_key(key, key_fname, pkey_passwd)

            if ca_enabled:
                # Start tracking only if the cert was issued by IPA CA
                # Retrieve IPA CA
                cdb = certs.CertDB(api.env.realm, nssdir=paths.IPA_NSSDB_DIR)
                ipa_ca_cert = cdb.get_cert_from_db(
                    get_ca_nickname(api.env.realm))
                # And compare with the CA which signed this certificate
                if ca_cert == ipa_ca_cert:
                    req_id = certmonger.start_tracking(
                        (cert_fname, key_fname),
                        pinfile=passwd_fname,
                        storage='FILE',
                        post_command=cmgr_post_command
                    )
                    return req_id
        except RuntimeError as e:
            raise admintool.ScriptError(str(e))
        return None

    def check_chain(self, pkcs12_filename, pkcs12_pin, nssdb):
        # create a temp nssdb
        with NSSDatabase() as tempnssdb:
            tempnssdb.create_db()

            # import the PKCS12 file, then delete all CA certificates
            # this leaves only the server certs in the temp db
            tempnssdb.import_pkcs12(pkcs12_filename, pkcs12_pin)
            for nickname, flags in tempnssdb.list_certs():
                if not flags.has_key:
                    while tempnssdb.has_nickname(nickname):
                        tempnssdb.delete_cert(nickname)

            # import all the CA certs from nssdb into the temp db
            for nickname, flags in nssdb.list_certs():
                if not flags.has_key:
                    cert = nssdb.get_cert_from_db(nickname)
                    tempnssdb.add_cert(cert, nickname, flags)

            # now get the server certs from tempnssdb and check their validity
            try:
                for nick, flags in tempnssdb.find_server_certs():
                    tempnssdb.verify_server_cert_validity(nick, api.env.host)
            except ValueError as e:
                raise admintool.ScriptError(
                    "Peer's certificate issuer is not trusted (%s). "
                    "Please run ipa-cacert-manage install and ipa-certupdate "
                    "to install the CA certificate." % str(e))

    def import_cert(self, dirname, pkcs12_passwd, old_cert, principal, command):
        pkcs12_file, pin, ca_cert = installutils.load_pkcs12(
            cert_files=self.args,
            key_password=pkcs12_passwd,
            key_nickname=self.options.cert_name,
            ca_cert_files=[paths.IPA_CA_CRT],
            host_name=api.env.host)

        dirname = os.path.normpath(dirname)
        cdb = certs.CertDB(api.env.realm, nssdir=dirname)

        # Check that the ca_cert is known and trusted
        self.check_chain(pkcs12_file.name, pin, cdb)

        try:
            ca_enabled = api.Command.ca_is_enabled()['result']
            if ca_enabled:
                cdb.untrack_server_cert(old_cert)

            cdb.delete_cert(old_cert)
            prevs = cdb.find_server_certs()
            cdb.import_pkcs12(pkcs12_file.name, pin)
            news = cdb.find_server_certs()
            server_certs = [item for item in news if item not in prevs]
            server_cert = server_certs[0][0]

            if ca_enabled:
                # Start tracking only if the cert was issued by IPA CA
                # Retrieve IPA CA
                ipa_ca_cert = cdb.get_cert_from_db(
                    get_ca_nickname(api.env.realm))
                # And compare with the CA which signed this certificate
                if ca_cert == ipa_ca_cert:
                    cdb.track_server_cert(server_cert,
                                          principal,
                                          cdb.passwd_fname,
                                          command)
        except RuntimeError as e:
            raise admintool.ScriptError(str(e))

        return server_cert



# Authors:
#   Petr Vobornik <pvoborni@redhat.com>
#
# Copyright (C) 2013  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Cert tests
"""

from ipatests.test_webui.crypto_utils import generate_csr
from ipatests.test_webui.ui_driver import UI_driver
from ipatests.test_webui.ui_driver import screenshot
from datetime import date, timedelta
import pytest

ENTITY = 'cert'

ERR_SPACE = "invalid '{}': Leading and trailing spaces are not allowed"
ERR_MUST_INTEGER = "invalid '{}': must be an integer"
LEAST_SERIAL = "invalid '{}': must be at least 0"
INV_DATE = ("invalid '{}': does not match any of accepted formats: "
            "%Y%m%d%H%M%SZ, %Y-%m-%dT%H:%M:%SZ, %Y-%m-%dT%H:%MZ, "
            "%Y-%m-%dZ, %Y-%m-%d %H:%M:%SZ, %Y-%m-%d %H:%MZ")


def search_pkey(self, pkey):
    search_field_s = '.search-filter input[name=filter]'
    self.fill_text(search_field_s, pkey)
    self.action_button_click('find', parent=None)
    self.wait_for_request(n=2)


def check_option_negative(self, date, option):
    self.navigate_to_entity(ENTITY)
    self.select('select[name=search_option]', option)
    search_pkey(self, date)
    self.assert_last_error_dialog(INV_DATE.format(option))
    self.close_all_dialogs()


def check_space_error(self, string, option):
    self.navigate_to_entity(ENTITY)
    self.select('select[name=search_option]', option)
    search_pkey(self, string)
    self.assert_last_error_dialog(ERR_SPACE.format(option))
    self.close_all_dialogs()


def check_integer(self, string, option):
    """
    Method to check if provided value is integer.
    If not check for error dialog
    """
    self.navigate_to_entity(ENTITY)
    self.select('select[name=search_option]', option)
    search_pkey(self, string)
    self.assert_last_error_dialog(ERR_MUST_INTEGER.format(option))
    self.close_all_dialogs()


def check_minimum_serial(self, serial, option):
    self.navigate_to_entity(ENTITY)
    self.select('select[name=search_option]', option)
    search_pkey(self, serial)
    self.assert_last_error_dialog(LEAST_SERIAL.format(option))
    self.close_all_dialogs()


@pytest.mark.tier1
class test_cert(UI_driver):

    def setup(self, *args, **kwargs):
        super(test_cert, self).setup(*args, **kwargs)

        if not self.has_ca():
            self.skip('CA not configured')

    def _add_and_revoke_cert(self, reason='1'):
        hostname = self.config.get('ipa_server')
        csr = generate_csr(hostname)

        self.navigate_to_entity(ENTITY)
        self.facet_button_click('request_cert')
        self.fill_textbox('principal', 'HTTP/{}'.format(hostname))
        self.check_option('add', 'checked')
        self.fill_textarea('csr', csr)
        self.dialog_button_click('issue')
        self.assert_notification(assert_text='Certificate requested')
        self.navigate_to_entity(ENTITY)
        rows = self.get_rows()
        cert = rows[-1]

        self.navigate_to_row_record(cert)
        self.action_list_action('revoke_cert', False)
        self.select('select[name=revocation_reason]', reason)
        self.dialog_button_click('ok')
        self.navigate_to_entity(ENTITY)

        return cert

    @screenshot
    def test_read(self):
        """
        Basic read: cert

        Certs don't have standard mod, add and delete methods.
        """
        self.init_app()
        self.navigate_to_entity(ENTITY)
        rows = self.get_rows()
        self.navigate_to_row_record(rows[0])
        self.navigate_by_breadcrumb("Certificates")

    @screenshot
    def test_search_subject(self):
        """
        Try to search certificate by subject
        """
        self.init_app()
        self.navigate_to_entity(ENTITY)
        self.select('select[name=search_option]', 'subject')
        search_pkey(self, 'Certificate Authority')
        rows = self.get_rows()
        assert len(rows) != 0

        # try to search non-existent subject
        self.navigate_to_entity(ENTITY)
        self.select('select[name=search_option]', 'subject')
        search_pkey(self, 'nonexistent')
        rows = self.get_rows()
        assert len(rows) == 0

        # try to search subject with speacial char
        self.navigate_to_entity(ENTITY)
        self.select('select[name=search_option]', 'subject')
        search_pkey(self, '<,>.?/')
        rows = self.get_rows()
        assert len(rows) == 0

        # try to search subject with leading space
        check_space_error(self, ' Certificate Authority', 'subject')

        # try to search subject with trailing space
        check_space_error(self, 'Certificate Authority ', 'subject')

    @screenshot
    def test_search_revocation_reason(self):
        """
        Try to search certificates by revocation reason
        """
        self.init_app()

        # revoke new certificate
        self._add_and_revoke_cert()

        # search cert by revocation reason
        self.select('select[name=search_option]', 'revocation_reason')
        search_pkey(self, '1')
        rows = self.get_rows()
        assert len(rows) != 0

        # search cert by string.
        check_integer(self, 'nonexistent', 'revocation_reason')

        # search cert by special char
        check_integer(self, '<,>.?/', 'revocation_reason')

        # search revocation reason negative Number.
        self.navigate_to_entity(ENTITY)
        self.select('select[name=search_option]', 'revocation_reason')
        search_pkey(self, '-1')
        rows = self.get_rows()
        assert len(rows) == 0

        # valid revocation reason can be value from 0 to 10
        # try revocation reason as other than valid value
        self.navigate_to_entity(ENTITY)
        self.select('select[name=search_option]', 'revocation_reason')
        search_pkey(self, '11')
        rows = self.get_rows()
        assert len(rows) == 0

    @screenshot
    def test_search_minimum_serial(self):
        """
        Try to search cert using minimum serial number option
        """
        self.init_app()
        self.navigate_to_entity(ENTITY)
        self.select('select[name=search_option]', 'min_serial_number')
        search_pkey(self, '1')
        rows = self.get_rows()
        assert len(rows) != 0

        # try search using string
        check_integer(self, 'nonexistent', 'min_serial_number')

        # try searching using -1
        check_minimum_serial(self, '-1', 'min_serial_number')

        # try using higher value than no. of certs present
        self.navigate_to_entity(ENTITY)
        self.select('select[name=search_option]', 'min_serial_number')
        search_pkey(self, '99')
        rows = self.get_rows()
        assert len(rows) == 0

    @screenshot
    def test_search_maximum_serial(self):
        """
        Try to search cert using maximum serial number option
        """
        self.init_app()
        self.navigate_to_entity(ENTITY)
        self.select('select[name=search_option]', 'max_serial_number')
        search_pkey(self, '2')
        rows = self.get_rows()
        assert len(rows) == 2

        # try to search using string
        check_integer(self, 'nonexisting', 'max_serial_number')

        # try to search using -1
        check_minimum_serial(self, '-1', 'max_serial_number')

    @screenshot
    def test_search_valid_not_after_from(self):
        """
        Try to search cert using valid not after from option
        """
        today = date.today()
        self.init_app()

        # revoke new certificate
        self._add_and_revoke_cert()

        self.select('select[name=search_option]', 'validnotafter_from')
        search_pkey(self, str(today))
        rows = self.get_rows()
        assert len(rows) != 0

        # try to search with string
        check_option_negative(self, 'nonexistent', 'validnotafter_from')

        # try to search using invalid date
        check_option_negative(self, '2018-02-30', 'validnotafter_from')

        # try to search using date beyond
        self.navigate_to_entity(ENTITY)
        self.select('select[name=search_option]', 'validnotafter_from')
        search_pkey(self, str(today + timedelta(weeks=52 * 30)))
        rows = self.get_rows()
        assert len(rows) == 0

        # try to search using leading space
        check_option_negative(self,
                              ' {}'.format(str(today)),
                              'validnotafter_from')

        # try to search trailing space
        check_option_negative(self,
                              '{} '.format(str(today)),
                              'validnotafter_from')

    @screenshot
    def test_search_valid_not_after_to(self):
        """
        Try to search cert using valid not after to option
        """
        today = date.today()
        self.init_app()

        # revoke new certificate
        self._add_and_revoke_cert()

        self.select('select[name=search_option]', 'validnotafter_to')
        search_pkey(self, str(today + timedelta(weeks=52 * 30)))
        rows = self.get_rows()
        assert len(rows) != 0

        # try to search with string
        check_option_negative(self, 'nonexistent', 'validnotafter_to')

        # try to search using invalid date
        check_option_negative(self, '2018-02-30', 'validnotafter_to')

        # try to search using date ago
        self.navigate_to_entity(ENTITY)
        self.select('select[name=search_option]', 'validnotafter_to')
        search_pkey(self, str(today - timedelta(weeks=52 * 10)))
        rows = self.get_rows()
        assert len(rows) == 0

        # try to search with leading space
        check_option_negative(self,
                              ' {}'.format(str(today)),
                              'validnotafter_to')

        # try to search with trailing space
        check_option_negative(self,
                              '{} '.format(str(today)),
                              'validnotafter_to')

    @screenshot
    def test_search_valid_not_before_from(self):
        """
        Try to search cert using valid not before from option
        """
        today = date.today()
        self.init_app()

        # revoke new certificate
        self._add_and_revoke_cert()

        self.select('select[name=search_option]', 'validnotbefore_from')
        search_pkey(self, str(today))
        rows = self.get_rows()
        assert len(rows) != 0

        # try to search with string
        check_option_negative(self, 'nonexistent', 'validnotafter_from')

        # try to search using invalid date
        check_option_negative(self, '2018-02-30', 'validnotafter_from')

        # try to search using current beyond
        self.navigate_to_entity(ENTITY)
        self.select('select[name=search_option]', 'validnotbefore_from')
        search_pkey(self, str(today + timedelta(weeks=52 * 30)))
        rows = self.get_rows()
        assert len(rows) == 0

        # try to search with leading space
        check_option_negative(self,
                              ' {}'.format(str(today)),
                              'validnotafter_from')

        # try to search with trailing space
        check_option_negative(self,
                              '{} '.format(str(today)),
                              'validnotafter_from')

    @screenshot
    def test_search_valid_not_before_to(self):
        """
        Try to search cert using valid not before to option
        """
        today = date.today()
        self.init_app()

        # revoke new certificate
        self._add_and_revoke_cert()

        self.select('select[name=search_option]', 'validnotbefore_to')
        search_pkey(self, str(today + timedelta(weeks=52 * 30)))
        rows = self.get_rows()
        assert len(rows) != 0

        # try to search with string
        check_option_negative(self, 'nonexistent', 'validnotafter_from')

        # try to search using invalid date
        check_option_negative(self, '2018-02-30', 'validnotafter_from')

        # try to search using date ago
        self.navigate_to_entity(ENTITY)
        self.select('select[name=search_option]', 'validnotbefore_to')
        search_pkey(self, str(today - timedelta(weeks=52 * 10)))
        rows = self.get_rows()
        assert len(rows) == 0

        # try to search with leading space
        check_option_negative(self,
                              ' {}'.format(str(today)),
                              'validnotafter_from')

        # try to search with trailing space
        check_option_negative(self,
                              '{} '.format(str(today)),
                              'validnotafter_from')

    @screenshot
    def test_search_issued_on_from(self):
        """
        Try to search cert using issued on from option
        """
        today = date.today()
        self.init_app()

        # revoke new certificate
        self._add_and_revoke_cert()

        self.select('select[name=search_option]', 'issuedon_from')
        search_pkey(self, str(today))
        rows = self.get_rows()
        assert len(rows) != 0

        # try to search with string
        check_option_negative(self, 'nonexistent', 'issuedon_from')

        # try to search using invalid date
        check_option_negative(self, '2018-02-30', 'issuedon_from')

        # try to search using date beyond
        self.navigate_to_entity(ENTITY)
        self.select('select[name=search_option]', 'issuedon_from')
        search_pkey(self, str(today + timedelta(weeks=52 * 30)))
        rows = self.get_rows()
        assert len(rows) == 0

        # try to search with leading space
        check_option_negative(self, ' {}'.format(str(today)), 'issuedon_from')

        # try to search with trailing space
        check_option_negative(self, '{} '.format(str(today)), 'issuedon_from')

    @screenshot
    def test_search_issued_on_to(self):
        """
        Try to search cert using issued on to option
        """
        today = date.today()
        self.init_app()

        # revoke new certificate
        self._add_and_revoke_cert()

        self.select('select[name=search_option]', 'issuedon_to')
        search_pkey(self, str(today))
        rows = self.get_rows()
        assert len(rows) != 0

        # try to search with string
        check_option_negative(self, 'nonexistent', 'issuedon_to')

        # try to search using invalid date
        check_option_negative(self, '2018-02-30', 'issuedon_to')

        # try to search using date ago
        self.navigate_to_entity(ENTITY)
        self.select('select[name=search_option]', 'issuedon_to')
        search_pkey(self, str(today - timedelta(weeks=52 * 10)))
        rows = self.get_rows()
        assert len(rows) == 0

        # try to search with leading space
        check_option_negative(self, ' {}'.format(str(today)), 'issuedon_to')

        # try to search with trailing space
        check_option_negative(self, '{} '.format(str(today)), 'issuedon_to')

    @screenshot
    def test_search_revoked_on_from(self):
        """
        Try to search cert using revoked on from option
        """
        today = date.today()
        self.init_app()

        # revoke new certificate
        self._add_and_revoke_cert()

        self.select('select[name=search_option]', 'revokedon_from')
        search_pkey(self, str(today))
        rows = self.get_rows()
        assert len(rows) != 0

        # try to search with string
        check_option_negative(self, 'nonexistent', 'revokedon_from')

        # try to search using invalid date
        check_option_negative(self, '2018-02-30', 'revokedon_from')

        # try to search using date beyond
        self.navigate_to_entity(ENTITY)
        self.select('select[name=search_option]', 'revokedon_from')
        search_pkey(self, str(today + timedelta(weeks=52 * 30)))
        rows = self.get_rows()
        assert len(rows) == 0

        # try to search with leading space
        check_option_negative(self, ' {}'.format(str(today)), 'revokedon_from')

        # try to search with trailing space
        check_option_negative(self, '{} '.format(str(today)), 'revokedon_from')

    @screenshot
    def test_search_revoked_on_to(self):
        """
        Try to search cert using revoked on to option
        """
        today = date.today()
        self.init_app()

        # revoke new certificate
        self._add_and_revoke_cert()

        self.select('select[name=search_option]', 'revokedon_to')
        search_pkey(self, str(today))
        rows = self.get_rows()
        assert len(rows) != 0

        # try to search with string
        check_option_negative(self, 'nonexistent', 'revokedon_to')

        # try to search using invalid date
        check_option_negative(self, '2018-02-30', 'revokedon_to')

        # try to search using date ago
        self.navigate_to_entity(ENTITY)
        self.select('select[name=search_option]', 'revokedon_to')
        search_pkey(self, str(today - timedelta(weeks=52 * 10)))
        rows = self.get_rows()
        assert len(rows) == 0

        # try to search with leading space
        check_option_negative(self, ' {}'.format(str(today)), 'revokedon_to')

        # try to search with trailing space
        check_option_negative(self, '{} '.format(str(today)), 'revokedon_to')



#
# Copyright (C) 2016  FreeIPA Contributors see COPYING for license
#

# pylint: disable=unused-import
import six

from . import Command, Method, Object
from ipalib import api, parameters, output
from ipalib.parameters import DefaultFrom
from ipalib.plugable import Registry
from ipalib.text import _
from ipapython.dn import DN
from ipapython.dnsutil import DNSName

if six.PY3:
    unicode = str

__doc__ = _("""
Topology

Management of a replication topology.

Requires minimum domain level 1.
""")

register = Registry()


@register()
class topologysegment(Object):
    takes_params = (
        parameters.Str(
            'cn',
            primary_key=True,
            label=_(u'Segment name'),
            doc=_(u'Arbitrary string identifying the segment'),
        ),
        parameters.Str(
            'iparepltoposegmentleftnode',
            label=_(u'Left node'),
            doc=_(u'Left replication node - an IPA server'),
        ),
        parameters.Str(
            'iparepltoposegmentrightnode',
            label=_(u'Right node'),
            doc=_(u'Right replication node - an IPA server'),
        ),
        parameters.Str(
            'iparepltoposegmentdirection',
            label=_(u'Connectivity'),
            doc=_(u'Direction of replication between left and right replication node'),
        ),
        parameters.Str(
            'nsds5replicastripattrs',
            required=False,
            label=_(u'Attributes to strip'),
            doc=_(u'A space separated list of attributes which are removed from replication updates.'),
        ),
        parameters.Str(
            'nsds5replicatedattributelist',
            required=False,
            label=_(u'Attributes to replicate'),
            doc=_(u'Attributes that are not replicated to a consumer server during a fractional update. E.g., `(objectclass=*) $ EXCLUDE accountlockout memberof'),
        ),
        parameters.Str(
            'nsds5replicatedattributelisttotal',
            required=False,
            label=_(u'Attributes for total update'),
            doc=_(u'Attributes that are not replicated to a consumer server during a total update. E.g. (objectclass=*) $ EXCLUDE accountlockout'),
        ),
        parameters.Int(
            'nsds5replicatimeout',
            required=False,
            label=_(u'Session timeout'),
            doc=_(u'Number of seconds outbound LDAP operations waits for a response from the remote replica before timing out and failing'),
        ),
        parameters.Str(
            'nsds5replicaenabled',
            required=False,
            label=_(u'Replication agreement enabled'),
            doc=_(u'Whether a replication agreement is active, meaning whether replication is occurring per that agreement'),
        ),
    )


@register()
class topologysuffix(Object):
    takes_params = (
        parameters.Str(
            'cn',
            primary_key=True,
            label=_(u'Suffix name'),
        ),
        parameters.Str(
            'iparepltopoconfroot',
            label=_(u'LDAP suffix to be managed'),
        ),
    )


@register()
class topologysegment_add(Method):
    __doc__ = _("Add a new segment.")

    NO_CLI = True

    takes_args = (
        parameters.Str(
            'topologysuffixcn',
            cli_name='topologysuffix',
            label=_(u'Suffix name'),
        ),
        parameters.Str(
            'cn',
            cli_name='name',
            label=_(u'Segment name'),
            doc=_(u'Arbitrary string identifying the segment'),
            default_from=DefaultFrom(lambda iparepltoposegmentleftnode, iparepltoposegmentrightnode: None, 'iparepltoposegmentleftnode', 'iparepltoposegmentrightnode'),
            # FIXME:
            # lambda iparepltoposegmentleftnode, iparepltoposegmentrightnode:
            no_convert=True,
        ),
    )
    takes_options = (
        parameters.Str(
            'iparepltoposegmentleftnode',
            cli_name='leftnode',
            label=_(u'Left node'),
            doc=_(u'Left replication node - an IPA server'),
            no_convert=True,
        ),
        parameters.Str(
            'iparepltoposegmentrightnode',
            cli_name='rightnode',
            label=_(u'Right node'),
            doc=_(u'Right replication node - an IPA server'),
            no_convert=True,
        ),
        parameters.Str(
            'iparepltoposegmentdirection',
            cli_name='direction',
            cli_metavar="['both', 'left-right', 'right-left']",
            label=_(u'Connectivity'),
            doc=_(u'Direction of replication between left and right replication node'),
            exclude=('cli', 'webui'),
            default=u'both',
        ),
        parameters.Str(
            'nsds5replicastripattrs',
            required=False,
            cli_name='stripattrs',
            label=_(u'Attributes to strip'),
            doc=_(u'A space separated list of attributes which are removed from replication updates.'),
            no_convert=True,
        ),
        parameters.Str(
            'nsds5replicatedattributelist',
            required=False,
            cli_name='replattrs',
            label=_(u'Attributes to replicate'),
            doc=_(u'Attributes that are not replicated to a consumer server during a fractional update. E.g., `(objectclass=*) $ EXCLUDE accountlockout memberof'),
        ),
        parameters.Str(
            'nsds5replicatedattributelisttotal',
            required=False,
            cli_name='replattrstotal',
            label=_(u'Attributes for total update'),
            doc=_(u'Attributes that are not replicated to a consumer server during a total update. E.g. (objectclass=*) $ EXCLUDE accountlockout'),
        ),
        parameters.Int(
            'nsds5replicatimeout',
            required=False,
            cli_name='timeout',
            label=_(u'Session timeout'),
            doc=_(u'Number of seconds outbound LDAP operations waits for a response from the remote replica before timing out and failing'),
        ),
        parameters.Str(
            'nsds5replicaenabled',
            required=False,
            cli_name='enabled',
            cli_metavar="['on', 'off']",
            label=_(u'Replication agreement enabled'),
            doc=_(u'Whether a replication agreement is active, meaning whether replication is occurring per that agreement'),
            exclude=('cli', 'webui'),
        ),
        parameters.Str(
            'setattr',
            required=False,
            multivalue=True,
            doc=_(u'Set an attribute to a name/value pair. Format is attr=value.\nFor multi-valued attributes, the command replaces the values already present.'),
            exclude=('webui',),
        ),
        parameters.Str(
            'addattr',
            required=False,
            multivalue=True,
            doc=_(u'Add an attribute/value pair. Format is attr=value. The attribute\nmust be part of the schema.'),
            exclude=('webui',),
        ),
        parameters.Flag(
            'all',
            doc=_(u'Retrieve and print all attributes from the server. Affects command output.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'raw',
            doc=_(u'Print entries as stored on the server. Only affects output format.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.Entry(
            'result',
        ),
        output.PrimaryKey(
            'value',
            doc=_(u"The primary_key value of the entry, e.g. 'jdoe' for a user"),
        ),
    )


@register()
class topologysegment_del(Method):
    __doc__ = _("Delete a segment.")

    NO_CLI = True

    takes_args = (
        parameters.Str(
            'topologysuffixcn',
            cli_name='topologysuffix',
            label=_(u'Suffix name'),
        ),
        parameters.Str(
            'cn',
            multivalue=True,
            cli_name='name',
            label=_(u'Segment name'),
            doc=_(u'Arbitrary string identifying the segment'),
            default_from=DefaultFrom(lambda iparepltoposegmentleftnode, iparepltoposegmentrightnode: None, 'iparepltoposegmentleftnode', 'iparepltoposegmentrightnode'),
            # FIXME:
            # lambda iparepltoposegmentleftnode, iparepltoposegmentrightnode:
            no_convert=True,
        ),
    )
    takes_options = (
        parameters.Flag(
            'continue',
            doc=_(u"Continuous mode: Don't stop on errors."),
            default=False,
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.Output(
            'result',
            dict,
            doc=_(u'List of deletions that failed'),
        ),
        output.ListOfPrimaryKeys(
            'value',
        ),
    )


@register()
class topologysegment_find(Method):
    __doc__ = _("Search for topology segments.")

    NO_CLI = True

    takes_args = (
        parameters.Str(
            'topologysuffixcn',
            cli_name='topologysuffix',
            label=_(u'Suffix name'),
        ),
        parameters.Str(
            'criteria',
            required=False,
            doc=_(u'A string searched in all relevant object attributes'),
        ),
    )
    takes_options = (
        parameters.Str(
            'cn',
            required=False,
            cli_name='name',
            label=_(u'Segment name'),
            doc=_(u'Arbitrary string identifying the segment'),
            default_from=DefaultFrom(lambda iparepltoposegmentleftnode, iparepltoposegmentrightnode: None, 'iparepltoposegmentleftnode', 'iparepltoposegmentrightnode'),
            # FIXME:
            # lambda iparepltoposegmentleftnode, iparepltoposegmentrightnode:
            no_convert=True,
        ),
        parameters.Str(
            'iparepltoposegmentleftnode',
            required=False,
            cli_name='leftnode',
            label=_(u'Left node'),
            doc=_(u'Left replication node - an IPA server'),
            no_convert=True,
        ),
        parameters.Str(
            'iparepltoposegmentrightnode',
            required=False,
            cli_name='rightnode',
            label=_(u'Right node'),
            doc=_(u'Right replication node - an IPA server'),
            no_convert=True,
        ),
        parameters.Str(
            'iparepltoposegmentdirection',
            required=False,
            cli_name='direction',
            cli_metavar="['both', 'left-right', 'right-left']",
            label=_(u'Connectivity'),
            doc=_(u'Direction of replication between left and right replication node'),
            exclude=('cli', 'webui'),
            default=u'both',
        ),
        parameters.Str(
            'nsds5replicastripattrs',
            required=False,
            cli_name='stripattrs',
            label=_(u'Attributes to strip'),
            doc=_(u'A space separated list of attributes which are removed from replication updates.'),
            no_convert=True,
        ),
        parameters.Str(
            'nsds5replicatedattributelist',
            required=False,
            cli_name='replattrs',
            label=_(u'Attributes to replicate'),
            doc=_(u'Attributes that are not replicated to a consumer server during a fractional update. E.g., `(objectclass=*) $ EXCLUDE accountlockout memberof'),
        ),
        parameters.Str(
            'nsds5replicatedattributelisttotal',
            required=False,
            cli_name='replattrstotal',
            label=_(u'Attributes for total update'),
            doc=_(u'Attributes that are not replicated to a consumer server during a total update. E.g. (objectclass=*) $ EXCLUDE accountlockout'),
        ),
        parameters.Int(
            'nsds5replicatimeout',
            required=False,
            cli_name='timeout',
            label=_(u'Session timeout'),
            doc=_(u'Number of seconds outbound LDAP operations waits for a response from the remote replica before timing out and failing'),
        ),
        parameters.Str(
            'nsds5replicaenabled',
            required=False,
            cli_name='enabled',
            cli_metavar="['on', 'off']",
            label=_(u'Replication agreement enabled'),
            doc=_(u'Whether a replication agreement is active, meaning whether replication is occurring per that agreement'),
            exclude=('cli', 'webui'),
        ),
        parameters.Int(
            'timelimit',
            required=False,
            label=_(u'Time Limit'),
            doc=_(u'Time limit of search in seconds (0 is unlimited)'),
        ),
        parameters.Int(
            'sizelimit',
            required=False,
            label=_(u'Size Limit'),
            doc=_(u'Maximum number of entries returned (0 is unlimited)'),
        ),
        parameters.Flag(
            'all',
            doc=_(u'Retrieve and print all attributes from the server. Affects command output.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'raw',
            doc=_(u'Print entries as stored on the server. Only affects output format.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'pkey_only',
            required=False,
            label=_(u'Primary key only'),
            doc=_(u'Results should contain primary key attribute only ("name")'),
            default=False,
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.ListOfEntries(
            'result',
        ),
        output.Output(
            'count',
            int,
            doc=_(u'Number of entries returned'),
        ),
        output.Output(
            'truncated',
            bool,
            doc=_(u'True if not all results were returned'),
        ),
    )


@register()
class topologysegment_mod(Method):
    __doc__ = _("Modify a segment.")

    NO_CLI = True

    takes_args = (
        parameters.Str(
            'topologysuffixcn',
            cli_name='topologysuffix',
            label=_(u'Suffix name'),
        ),
        parameters.Str(
            'cn',
            cli_name='name',
            label=_(u'Segment name'),
            doc=_(u'Arbitrary string identifying the segment'),
            default_from=DefaultFrom(lambda iparepltoposegmentleftnode, iparepltoposegmentrightnode: None, 'iparepltoposegmentleftnode', 'iparepltoposegmentrightnode'),
            # FIXME:
            # lambda iparepltoposegmentleftnode, iparepltoposegmentrightnode:
            no_convert=True,
        ),
    )
    takes_options = (
        parameters.Str(
            'nsds5replicastripattrs',
            required=False,
            cli_name='stripattrs',
            label=_(u'Attributes to strip'),
            doc=_(u'A space separated list of attributes which are removed from replication updates.'),
            no_convert=True,
        ),
        parameters.Str(
            'nsds5replicatedattributelist',
            required=False,
            cli_name='replattrs',
            label=_(u'Attributes to replicate'),
            doc=_(u'Attributes that are not replicated to a consumer server during a fractional update. E.g., `(objectclass=*) $ EXCLUDE accountlockout memberof'),
        ),
        parameters.Str(
            'nsds5replicatedattributelisttotal',
            required=False,
            cli_name='replattrstotal',
            label=_(u'Attributes for total update'),
            doc=_(u'Attributes that are not replicated to a consumer server during a total update. E.g. (objectclass=*) $ EXCLUDE accountlockout'),
        ),
        parameters.Int(
            'nsds5replicatimeout',
            required=False,
            cli_name='timeout',
            label=_(u'Session timeout'),
            doc=_(u'Number of seconds outbound LDAP operations waits for a response from the remote replica before timing out and failing'),
        ),
        parameters.Str(
            'nsds5replicaenabled',
            required=False,
            cli_name='enabled',
            cli_metavar="['on', 'off']",
            label=_(u'Replication agreement enabled'),
            doc=_(u'Whether a replication agreement is active, meaning whether replication is occurring per that agreement'),
            exclude=('cli', 'webui'),
        ),
        parameters.Str(
            'setattr',
            required=False,
            multivalue=True,
            doc=_(u'Set an attribute to a name/value pair. Format is attr=value.\nFor multi-valued attributes, the command replaces the values already present.'),
            exclude=('webui',),
        ),
        parameters.Str(
            'addattr',
            required=False,
            multivalue=True,
            doc=_(u'Add an attribute/value pair. Format is attr=value. The attribute\nmust be part of the schema.'),
            exclude=('webui',),
        ),
        parameters.Str(
            'delattr',
            required=False,
            multivalue=True,
            doc=_(u'Delete an attribute/value pair. The option will be evaluated\nlast, after all sets and adds.'),
            exclude=('webui',),
        ),
        parameters.Flag(
            'rights',
            label=_(u'Rights'),
            doc=_(u'Display the access rights of this entry (requires --all). See ipa man page for details.'),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'all',
            doc=_(u'Retrieve and print all attributes from the server. Affects command output.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'raw',
            doc=_(u'Print entries as stored on the server. Only affects output format.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.Entry(
            'result',
        ),
        output.PrimaryKey(
            'value',
            doc=_(u"The primary_key value of the entry, e.g. 'jdoe' for a user"),
        ),
    )


@register()
class topologysegment_reinitialize(Method):
    __doc__ = _("Request a full re-initialization of the node retrieving data from the other node.")

    NO_CLI = True

    takes_args = (
        parameters.Str(
            'topologysuffixcn',
            cli_name='topologysuffix',
            label=_(u'Suffix name'),
        ),
        parameters.Str(
            'cn',
            cli_name='name',
            label=_(u'Segment name'),
            doc=_(u'Arbitrary string identifying the segment'),
            default_from=DefaultFrom(lambda iparepltoposegmentleftnode, iparepltoposegmentrightnode: None, 'iparepltoposegmentleftnode', 'iparepltoposegmentrightnode'),
            # FIXME:
            # lambda iparepltoposegmentleftnode, iparepltoposegmentrightnode:
            no_convert=True,
        ),
    )
    takes_options = (
        parameters.Flag(
            'left',
            required=False,
            doc=_(u'Initialize left node'),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'right',
            required=False,
            doc=_(u'Initialize right node'),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'stop',
            required=False,
            doc=_(u'Stop already started refresh of chosen node(s)'),
            default=False,
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.Output(
            'result',
            bool,
            doc=_(u'True means the operation was successful'),
        ),
        output.PrimaryKey(
            'value',
            doc=_(u"The primary_key value of the entry, e.g. 'jdoe' for a user"),
        ),
    )


@register()
class topologysegment_show(Method):
    __doc__ = _("Display a segment.")

    NO_CLI = True

    takes_args = (
        parameters.Str(
            'topologysuffixcn',
            cli_name='topologysuffix',
            label=_(u'Suffix name'),
        ),
        parameters.Str(
            'cn',
            cli_name='name',
            label=_(u'Segment name'),
            doc=_(u'Arbitrary string identifying the segment'),
            default_from=DefaultFrom(lambda iparepltoposegmentleftnode, iparepltoposegmentrightnode: None, 'iparepltoposegmentleftnode', 'iparepltoposegmentrightnode'),
            # FIXME:
            # lambda iparepltoposegmentleftnode, iparepltoposegmentrightnode:
            no_convert=True,
        ),
    )
    takes_options = (
        parameters.Flag(
            'rights',
            label=_(u'Rights'),
            doc=_(u'Display the access rights of this entry (requires --all). See ipa man page for details.'),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'all',
            doc=_(u'Retrieve and print all attributes from the server. Affects command output.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'raw',
            doc=_(u'Print entries as stored on the server. Only affects output format.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.Entry(
            'result',
        ),
        output.PrimaryKey(
            'value',
            doc=_(u"The primary_key value of the entry, e.g. 'jdoe' for a user"),
        ),
    )


@register()
class topologysuffix_add(Method):
    __doc__ = _("Add a new topology suffix to be managed.")

    NO_CLI = True

    takes_args = (
        parameters.Str(
            'cn',
            cli_name='name',
            label=_(u'Suffix name'),
        ),
    )
    takes_options = (
        parameters.Str(
            'iparepltopoconfroot',
            cli_name='suffix',
            label=_(u'LDAP suffix to be managed'),
            no_convert=True,
        ),
        parameters.Str(
            'setattr',
            required=False,
            multivalue=True,
            doc=_(u'Set an attribute to a name/value pair. Format is attr=value.\nFor multi-valued attributes, the command replaces the values already present.'),
            exclude=('webui',),
        ),
        parameters.Str(
            'addattr',
            required=False,
            multivalue=True,
            doc=_(u'Add an attribute/value pair. Format is attr=value. The attribute\nmust be part of the schema.'),
            exclude=('webui',),
        ),
        parameters.Flag(
            'all',
            doc=_(u'Retrieve and print all attributes from the server. Affects command output.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'raw',
            doc=_(u'Print entries as stored on the server. Only affects output format.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.Entry(
            'result',
        ),
        output.PrimaryKey(
            'value',
            doc=_(u"The primary_key value of the entry, e.g. 'jdoe' for a user"),
        ),
    )


@register()
class topologysuffix_del(Method):
    __doc__ = _("Delete a topology suffix.")

    NO_CLI = True

    takes_args = (
        parameters.Str(
            'cn',
            multivalue=True,
            cli_name='name',
            label=_(u'Suffix name'),
        ),
    )
    takes_options = (
        parameters.Flag(
            'continue',
            doc=_(u"Continuous mode: Don't stop on errors."),
            default=False,
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.Output(
            'result',
            dict,
            doc=_(u'List of deletions that failed'),
        ),
        output.ListOfPrimaryKeys(
            'value',
        ),
    )


@register()
class topologysuffix_find(Method):
    __doc__ = _("Search for topology suffices.")

    NO_CLI = True

    takes_args = (
        parameters.Str(
            'criteria',
            required=False,
            doc=_(u'A string searched in all relevant object attributes'),
        ),
    )
    takes_options = (
        parameters.Str(
            'cn',
            required=False,
            cli_name='name',
            label=_(u'Suffix name'),
        ),
        parameters.Str(
            'iparepltopoconfroot',
            required=False,
            cli_name='suffix',
            label=_(u'LDAP suffix to be managed'),
            no_convert=True,
        ),
        parameters.Int(
            'timelimit',
            required=False,
            label=_(u'Time Limit'),
            doc=_(u'Time limit of search in seconds (0 is unlimited)'),
        ),
        parameters.Int(
            'sizelimit',
            required=False,
            label=_(u'Size Limit'),
            doc=_(u'Maximum number of entries returned (0 is unlimited)'),
        ),
        parameters.Flag(
            'all',
            doc=_(u'Retrieve and print all attributes from the server. Affects command output.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'raw',
            doc=_(u'Print entries as stored on the server. Only affects output format.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'pkey_only',
            required=False,
            label=_(u'Primary key only'),
            doc=_(u'Results should contain primary key attribute only ("name")'),
            default=False,
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.ListOfEntries(
            'result',
        ),
        output.Output(
            'count',
            int,
            doc=_(u'Number of entries returned'),
        ),
        output.Output(
            'truncated',
            bool,
            doc=_(u'True if not all results were returned'),
        ),
    )


@register()
class topologysuffix_mod(Method):
    __doc__ = _("Modify a topology suffix.")

    NO_CLI = True

    takes_args = (
        parameters.Str(
            'cn',
            cli_name='name',
            label=_(u'Suffix name'),
        ),
    )
    takes_options = (
        parameters.Str(
            'iparepltopoconfroot',
            required=False,
            cli_name='suffix',
            label=_(u'LDAP suffix to be managed'),
            no_convert=True,
        ),
        parameters.Str(
            'setattr',
            required=False,
            multivalue=True,
            doc=_(u'Set an attribute to a name/value pair. Format is attr=value.\nFor multi-valued attributes, the command replaces the values already present.'),
            exclude=('webui',),
        ),
        parameters.Str(
            'addattr',
            required=False,
            multivalue=True,
            doc=_(u'Add an attribute/value pair. Format is attr=value. The attribute\nmust be part of the schema.'),
            exclude=('webui',),
        ),
        parameters.Str(
            'delattr',
            required=False,
            multivalue=True,
            doc=_(u'Delete an attribute/value pair. The option will be evaluated\nlast, after all sets and adds.'),
            exclude=('webui',),
        ),
        parameters.Flag(
            'rights',
            label=_(u'Rights'),
            doc=_(u'Display the access rights of this entry (requires --all). See ipa man page for details.'),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'all',
            doc=_(u'Retrieve and print all attributes from the server. Affects command output.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'raw',
            doc=_(u'Print entries as stored on the server. Only affects output format.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.Entry(
            'result',
        ),
        output.PrimaryKey(
            'value',
            doc=_(u"The primary_key value of the entry, e.g. 'jdoe' for a user"),
        ),
    )


@register()
class topologysuffix_show(Method):
    __doc__ = _("Show managed suffix.")

    NO_CLI = True

    takes_args = (
        parameters.Str(
            'cn',
            cli_name='name',
            label=_(u'Suffix name'),
        ),
    )
    takes_options = (
        parameters.Flag(
            'rights',
            label=_(u'Rights'),
            doc=_(u'Display the access rights of this entry (requires --all). See ipa man page for details.'),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'all',
            doc=_(u'Retrieve and print all attributes from the server. Affects command output.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'raw',
            doc=_(u'Print entries as stored on the server. Only affects output format.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.Entry(
            'result',
        ),
        output.PrimaryKey(
            'value',
            doc=_(u"The primary_key value of the entry, e.g. 'jdoe' for a user"),
        ),
    )


@register()
class topologysuffix_verify(Method):
    __doc__ = _("""
Verify replication topology for suffix.

Checks done:
  1. check if a topology is not disconnected. In other words if there are
     replication paths between all servers.
  2. check if servers don't have more than the recommended number of
     replication agreements
    """)

    NO_CLI = True

    takes_args = (
        parameters.Str(
            'cn',
            cli_name='name',
            label=_(u'Suffix name'),
        ),
    )
    takes_options = (
    )
    has_output = (
        output.Output(
            'result',
        ),
    )



#
# Copyright (C) 2017  FreeIPA Contributors see COPYING for license
#

import pytest

from ipatests.test_integration.base import IntegrationTest
from ipatests.pytest_ipa.integration.tasks import clear_sssd_cache


test_data = []
for i in range(3):
    data = {
        'user': {
            'login': 'testuser_{}'.format(i),
            'first': 'Test_{}'.format(i),
            'last': 'User_{}'.format(i),
        },
        'netgroup': 'testgroup_{}'.format(i),
        'nested_netgroup': 'testgroup_{}'.format(i-1) if i > 0 else None
    }
    test_data.append(data)
    members = [d['user']['login'] for d in test_data]
    test_data[-1]['netgroup_nested_members'] = members


@pytest.fixture()
def three_netgroups(request):
    """Prepare basic netgroups with users"""

    for d in test_data:
        request.cls.master.run_command(['ipa', 'user-add', d['user']['login'],
                                        '--first', d['user']['first'],
                                        '--last', d['user']['last']],
                                       raiseonerr=False)

        request.cls.master.run_command(['ipa', 'netgroup-add', d['netgroup']],
                                       raiseonerr=False)

        user_opt = '--users={u[login]}'.format(u=d['user'])
        request.cls.master.run_command(['ipa', 'netgroup-add-member', user_opt,
                                        d['netgroup']], raiseonerr=False)

    def teardown_three_netgroups():
        """Clean basic netgroups with users"""
        for d in test_data:
            request.cls.master.run_command(['ipa', 'user-del',
                                            d['user']['login']],
                                           raiseonerr=False)

            request.cls.master.run_command(['ipa', 'netgroup-del',
                                            d['netgroup']],
                                           raiseonerr=False)

    request.addfinalizer(teardown_three_netgroups)


class TestNetgroups(IntegrationTest):
    """
    Test Netgroups
    """

    topology = 'line'

    def check_users_in_netgroups(self):
        """Check if users are in groups, no nested things"""
        master = self.master
        clear_sssd_cache(master)

        for d in test_data:
            result = master.run_command(['getent', 'passwd',
                                         d['user']['login']], raiseonerr=False)
            assert result.returncode == 0

            user = '{u[first]} {u[last]}'.format(u=d['user'])
            assert user in result.stdout_text

            result = master.run_command(['getent', 'netgroup',
                                         d['netgroup']], raiseonerr=False)
            assert result.returncode == 0

            netgroup = '(-,{},{})'.format(d['user']['login'],
                                          self.master.domain.name)
            assert netgroup in result.stdout_text

    def check_nested_netgroup_hierarchy(self):
        """Check if nested netgroups hierarchy is complete"""
        master = self.master
        clear_sssd_cache(master)

        for d in test_data:
            result = master.run_command(['getent', 'netgroup', d['netgroup']],
                                        raiseonerr=False)
            assert result.returncode == 0

            for member in d['netgroup_nested_members']:
                if not member:
                    continue

                netgroup = '(-,{},{})'.format(member, self.master.domain.name)
                assert netgroup in result.stdout_text

    def prepare_nested_netgroup_hierarchy(self):
        """Prepares nested netgroup hierarchy from basic netgroups"""
        for d in test_data:
            if not d['nested_netgroup']:
                continue

            netgroups_opt = '--netgroups={}'.format(d['nested_netgroup'])
            self.master.run_command(['ipa', 'netgroup-add-member',
                                     netgroups_opt, d['netgroup']])

    def test_add_nested_netgroup(self, three_netgroups):
        """Test of adding nested groups"""
        self.check_users_in_netgroups()
        self.prepare_nested_netgroup_hierarchy()
        self.check_nested_netgroup_hierarchy()

    def test_remove_nested_netgroup(self, three_netgroups):
        """Test of removing nested groups"""
        master = self.master

        trinity = ['(-,{},{})'.format(d['user']['login'],
                                      self.master.domain.name)
                   for d in test_data]

        self.check_users_in_netgroups()
        self.prepare_nested_netgroup_hierarchy()
        self.check_nested_netgroup_hierarchy()

        # Removing of testgroup_1 from testgroup_2
        netgroups_opt = '--netgroups={n[netgroup]}'.format(n=test_data[0])
        result = self.master.run_command(['ipa', 'netgroup-remove-member',
                                          netgroups_opt,
                                          test_data[1]['netgroup']],
                                         raiseonerr=False)
        assert result.returncode == 0
        clear_sssd_cache(master)

        result = master.run_command(['getent', 'netgroup',
                                     test_data[1]['netgroup']],
                                    raiseonerr=False)
        assert result.returncode == 0
        assert trinity[1] in result.stdout_text

        result = master.run_command(['getent', 'netgroup',
                                     test_data[2]['netgroup']],
                                    raiseonerr=False)
        assert result.returncode == 0
        assert trinity[0] not in result.stdout_text
        assert trinity[1] in result.stdout_text
        assert trinity[2] in result.stdout_text

        # Removing of testgroup_2 from testgroup_3
        netgroups_opt = '--netgroups={n[netgroup]}'.format(n=test_data[1])
        result = self.master.run_command(['ipa', 'netgroup-remove-member',
                                          netgroups_opt,
                                          test_data[2]['netgroup']],
                                         raiseonerr=False)
        assert result.returncode == 0
        clear_sssd_cache(master)

        result = master.run_command(['getent', 'netgroup',
                                     test_data[2]['netgroup']],
                                    raiseonerr=False)
        assert result.returncode == 0
        assert trinity[0] not in result.stdout_text
        assert trinity[1] not in result.stdout_text
        assert trinity[2] in result.stdout_text



#
# Copyright (C) 2016  FreeIPA Contributors see COPYING for license
#

# pylint: disable=unused-import
import six

from . import Command, Method, Object
from ipalib import api, parameters, output
from ipalib.parameters import DefaultFrom
from ipalib.plugable import Registry
from ipalib.text import _
from ipapython.dn import DN
from ipapython.dnsutil import DNSName

if six.PY3:
    unicode = str

__doc__ = _("""
ID Views

Manage ID Views

IPA allows to override certain properties of users and groups per each host.
This functionality is primarily used to allow migration from older systems or
other Identity Management solutions.
""")

register = Registry()


@register()
class idoverridegroup(Object):
    takes_params = (
        parameters.Str(
            'ipaanchoruuid',
            primary_key=True,
            label=_(u'Anchor to override'),
        ),
        parameters.Str(
            'description',
            required=False,
            label=_(u'Description'),
        ),
        parameters.Str(
            'cn',
            required=False,
            label=_(u'Group name'),
        ),
        parameters.Int(
            'gidnumber',
            required=False,
            label=_(u'GID'),
            doc=_(u'Group ID Number'),
        ),
    )


@register()
class idoverrideuser(Object):
    takes_params = (
        parameters.Str(
            'ipaanchoruuid',
            primary_key=True,
            label=_(u'Anchor to override'),
        ),
        parameters.Str(
            'description',
            required=False,
            label=_(u'Description'),
        ),
        parameters.Str(
            'uid',
            required=False,
            label=_(u'User login'),
        ),
        parameters.Int(
            'uidnumber',
            required=False,
            label=_(u'UID'),
            doc=_(u'User ID Number'),
        ),
        parameters.Str(
            'gecos',
            required=False,
            label=_(u'GECOS'),
        ),
        parameters.Int(
            'gidnumber',
            required=False,
            label=_(u'GID'),
            doc=_(u'Group ID Number'),
        ),
        parameters.Str(
            'homedirectory',
            required=False,
            label=_(u'Home directory'),
        ),
        parameters.Str(
            'loginshell',
            required=False,
            label=_(u'Login shell'),
        ),
        parameters.Str(
            'ipaoriginaluid',
            required=False,
            exclude=('cli', 'webui'),
        ),
        parameters.Str(
            'ipasshpubkey',
            required=False,
            multivalue=True,
            label=_(u'SSH public key'),
        ),
    )


@register()
class idview(Object):
    takes_params = (
        parameters.Str(
            'cn',
            primary_key=True,
            label=_(u'ID View Name'),
        ),
        parameters.Str(
            'description',
            required=False,
            label=_(u'Description'),
        ),
    )


@register()
class idoverridegroup_add(Method):
    __doc__ = _("Add a new Group ID override.")

    takes_args = (
        parameters.Str(
            'idviewcn',
            cli_name='idview',
            label=_(u'ID View Name'),
        ),
        parameters.Str(
            'ipaanchoruuid',
            cli_name='anchor',
            label=_(u'Anchor to override'),
        ),
    )
    takes_options = (
        parameters.Str(
            'description',
            required=False,
            cli_name='desc',
            label=_(u'Description'),
        ),
        parameters.Str(
            'cn',
            required=False,
            cli_name='group_name',
            label=_(u'Group name'),
            no_convert=True,
        ),
        parameters.Int(
            'gidnumber',
            required=False,
            cli_name='gid',
            label=_(u'GID'),
            doc=_(u'Group ID Number'),
        ),
        parameters.Str(
            'setattr',
            required=False,
            multivalue=True,
            doc=_(u'Set an attribute to a name/value pair. Format is attr=value.\nFor multi-valued attributes, the command replaces the values already present.'),
            exclude=('webui',),
        ),
        parameters.Str(
            'addattr',
            required=False,
            multivalue=True,
            doc=_(u'Add an attribute/value pair. Format is attr=value. The attribute\nmust be part of the schema.'),
            exclude=('webui',),
        ),
        parameters.Flag(
            'all',
            doc=_(u'Retrieve and print all attributes from the server. Affects command output.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'raw',
            doc=_(u'Print entries as stored on the server. Only affects output format.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.Entry(
            'result',
        ),
        output.PrimaryKey(
            'value',
            doc=_(u"The primary_key value of the entry, e.g. 'jdoe' for a user"),
        ),
    )


@register()
class idoverridegroup_del(Method):
    __doc__ = _("Delete an Group ID override.")

    takes_args = (
        parameters.Str(
            'idviewcn',
            cli_name='idview',
            label=_(u'ID View Name'),
        ),
        parameters.Str(
            'ipaanchoruuid',
            multivalue=True,
            cli_name='anchor',
            label=_(u'Anchor to override'),
        ),
    )
    takes_options = (
        parameters.Flag(
            'continue',
            doc=_(u"Continuous mode: Don't stop on errors."),
            default=False,
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.Output(
            'result',
            dict,
            doc=_(u'List of deletions that failed'),
        ),
        output.ListOfPrimaryKeys(
            'value',
        ),
    )


@register()
class idoverridegroup_find(Method):
    __doc__ = _("Search for an Group ID override.")

    takes_args = (
        parameters.Str(
            'idviewcn',
            cli_name='idview',
            label=_(u'ID View Name'),
        ),
        parameters.Str(
            'criteria',
            required=False,
            doc=_(u'A string searched in all relevant object attributes'),
        ),
    )
    takes_options = (
        parameters.Str(
            'ipaanchoruuid',
            required=False,
            cli_name='anchor',
            label=_(u'Anchor to override'),
        ),
        parameters.Str(
            'description',
            required=False,
            cli_name='desc',
            label=_(u'Description'),
        ),
        parameters.Str(
            'cn',
            required=False,
            cli_name='group_name',
            label=_(u'Group name'),
            no_convert=True,
        ),
        parameters.Int(
            'gidnumber',
            required=False,
            cli_name='gid',
            label=_(u'GID'),
            doc=_(u'Group ID Number'),
        ),
        parameters.Int(
            'timelimit',
            required=False,
            label=_(u'Time Limit'),
            doc=_(u'Time limit of search in seconds'),
        ),
        parameters.Int(
            'sizelimit',
            required=False,
            label=_(u'Size Limit'),
            doc=_(u'Maximum number of entries returned'),
        ),
        parameters.Flag(
            'all',
            doc=_(u'Retrieve and print all attributes from the server. Affects command output.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'raw',
            doc=_(u'Print entries as stored on the server. Only affects output format.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'pkey_only',
            required=False,
            label=_(u'Primary key only'),
            doc=_(u'Results should contain primary key attribute only ("anchor")'),
            default=False,
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.ListOfEntries(
            'result',
        ),
        output.Output(
            'count',
            int,
            doc=_(u'Number of entries returned'),
        ),
        output.Output(
            'truncated',
            bool,
            doc=_(u'True if not all results were returned'),
        ),
    )


@register()
class idoverridegroup_mod(Method):
    __doc__ = _("Modify an Group ID override.")

    takes_args = (
        parameters.Str(
            'idviewcn',
            cli_name='idview',
            label=_(u'ID View Name'),
        ),
        parameters.Str(
            'ipaanchoruuid',
            cli_name='anchor',
            label=_(u'Anchor to override'),
        ),
    )
    takes_options = (
        parameters.Str(
            'description',
            required=False,
            cli_name='desc',
            label=_(u'Description'),
        ),
        parameters.Str(
            'cn',
            required=False,
            cli_name='group_name',
            label=_(u'Group name'),
            no_convert=True,
        ),
        parameters.Int(
            'gidnumber',
            required=False,
            cli_name='gid',
            label=_(u'GID'),
            doc=_(u'Group ID Number'),
        ),
        parameters.Str(
            'setattr',
            required=False,
            multivalue=True,
            doc=_(u'Set an attribute to a name/value pair. Format is attr=value.\nFor multi-valued attributes, the command replaces the values already present.'),
            exclude=('webui',),
        ),
        parameters.Str(
            'addattr',
            required=False,
            multivalue=True,
            doc=_(u'Add an attribute/value pair. Format is attr=value. The attribute\nmust be part of the schema.'),
            exclude=('webui',),
        ),
        parameters.Str(
            'delattr',
            required=False,
            multivalue=True,
            doc=_(u'Delete an attribute/value pair. The option will be evaluated\nlast, after all sets and adds.'),
            exclude=('webui',),
        ),
        parameters.Flag(
            'rights',
            label=_(u'Rights'),
            doc=_(u'Display the access rights of this entry (requires --all). See ipa man page for details.'),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'all',
            doc=_(u'Retrieve and print all attributes from the server. Affects command output.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'raw',
            doc=_(u'Print entries as stored on the server. Only affects output format.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Str(
            'rename',
            required=False,
            label=_(u'Rename'),
            doc=_(u'Rename the Group ID override object'),
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.Entry(
            'result',
        ),
        output.PrimaryKey(
            'value',
            doc=_(u"The primary_key value of the entry, e.g. 'jdoe' for a user"),
        ),
    )


@register()
class idoverridegroup_show(Method):
    __doc__ = _("Display information about an Group ID override.")

    takes_args = (
        parameters.Str(
            'idviewcn',
            cli_name='idview',
            label=_(u'ID View Name'),
        ),
        parameters.Str(
            'ipaanchoruuid',
            cli_name='anchor',
            label=_(u'Anchor to override'),
        ),
    )
    takes_options = (
        parameters.Flag(
            'rights',
            label=_(u'Rights'),
            doc=_(u'Display the access rights of this entry (requires --all). See ipa man page for details.'),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'all',
            doc=_(u'Retrieve and print all attributes from the server. Affects command output.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'raw',
            doc=_(u'Print entries as stored on the server. Only affects output format.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.Entry(
            'result',
        ),
        output.PrimaryKey(
            'value',
            doc=_(u"The primary_key value of the entry, e.g. 'jdoe' for a user"),
        ),
    )


@register()
class idoverrideuser_add(Method):
    __doc__ = _("Add a new User ID override.")

    takes_args = (
        parameters.Str(
            'idviewcn',
            cli_name='idview',
            label=_(u'ID View Name'),
        ),
        parameters.Str(
            'ipaanchoruuid',
            cli_name='anchor',
            label=_(u'Anchor to override'),
        ),
    )
    takes_options = (
        parameters.Str(
            'description',
            required=False,
            cli_name='desc',
            label=_(u'Description'),
        ),
        parameters.Str(
            'uid',
            required=False,
            cli_name='login',
            label=_(u'User login'),
            no_convert=True,
        ),
        parameters.Int(
            'uidnumber',
            required=False,
            cli_name='uid',
            label=_(u'UID'),
            doc=_(u'User ID Number'),
        ),
        parameters.Str(
            'gecos',
            required=False,
            label=_(u'GECOS'),
        ),
        parameters.Int(
            'gidnumber',
            required=False,
            label=_(u'GID'),
            doc=_(u'Group ID Number'),
        ),
        parameters.Str(
            'homedirectory',
            required=False,
            cli_name='homedir',
            label=_(u'Home directory'),
        ),
        parameters.Str(
            'loginshell',
            required=False,
            cli_name='shell',
            label=_(u'Login shell'),
        ),
        parameters.Str(
            'ipaoriginaluid',
            required=False,
            exclude=('cli', 'webui'),
        ),
        parameters.Str(
            'ipasshpubkey',
            required=False,
            multivalue=True,
            cli_name='sshpubkey',
            label=_(u'SSH public key'),
            no_convert=True,
        ),
        parameters.Str(
            'setattr',
            required=False,
            multivalue=True,
            doc=_(u'Set an attribute to a name/value pair. Format is attr=value.\nFor multi-valued attributes, the command replaces the values already present.'),
            exclude=('webui',),
        ),
        parameters.Str(
            'addattr',
            required=False,
            multivalue=True,
            doc=_(u'Add an attribute/value pair. Format is attr=value. The attribute\nmust be part of the schema.'),
            exclude=('webui',),
        ),
        parameters.Flag(
            'all',
            doc=_(u'Retrieve and print all attributes from the server. Affects command output.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'raw',
            doc=_(u'Print entries as stored on the server. Only affects output format.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.Entry(
            'result',
        ),
        output.PrimaryKey(
            'value',
            doc=_(u"The primary_key value of the entry, e.g. 'jdoe' for a user"),
        ),
    )


@register()
class idoverrideuser_del(Method):
    __doc__ = _("Delete an User ID override.")

    takes_args = (
        parameters.Str(
            'idviewcn',
            cli_name='idview',
            label=_(u'ID View Name'),
        ),
        parameters.Str(
            'ipaanchoruuid',
            multivalue=True,
            cli_name='anchor',
            label=_(u'Anchor to override'),
        ),
    )
    takes_options = (
        parameters.Flag(
            'continue',
            doc=_(u"Continuous mode: Don't stop on errors."),
            default=False,
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.Output(
            'result',
            dict,
            doc=_(u'List of deletions that failed'),
        ),
        output.ListOfPrimaryKeys(
            'value',
        ),
    )


@register()
class idoverrideuser_find(Method):
    __doc__ = _("Search for an User ID override.")

    takes_args = (
        parameters.Str(
            'idviewcn',
            cli_name='idview',
            label=_(u'ID View Name'),
        ),
        parameters.Str(
            'criteria',
            required=False,
            doc=_(u'A string searched in all relevant object attributes'),
        ),
    )
    takes_options = (
        parameters.Str(
            'ipaanchoruuid',
            required=False,
            cli_name='anchor',
            label=_(u'Anchor to override'),
        ),
        parameters.Str(
            'description',
            required=False,
            cli_name='desc',
            label=_(u'Description'),
        ),
        parameters.Str(
            'uid',
            required=False,
            cli_name='login',
            label=_(u'User login'),
            no_convert=True,
        ),
        parameters.Int(
            'uidnumber',
            required=False,
            cli_name='uid',
            label=_(u'UID'),
            doc=_(u'User ID Number'),
        ),
        parameters.Str(
            'gecos',
            required=False,
            label=_(u'GECOS'),
        ),
        parameters.Int(
            'gidnumber',
            required=False,
            label=_(u'GID'),
            doc=_(u'Group ID Number'),
        ),
        parameters.Str(
            'homedirectory',
            required=False,
            cli_name='homedir',
            label=_(u'Home directory'),
        ),
        parameters.Str(
            'loginshell',
            required=False,
            cli_name='shell',
            label=_(u'Login shell'),
        ),
        parameters.Str(
            'ipaoriginaluid',
            required=False,
            exclude=('cli', 'webui'),
        ),
        parameters.Int(
            'timelimit',
            required=False,
            label=_(u'Time Limit'),
            doc=_(u'Time limit of search in seconds'),
        ),
        parameters.Int(
            'sizelimit',
            required=False,
            label=_(u'Size Limit'),
            doc=_(u'Maximum number of entries returned'),
        ),
        parameters.Flag(
            'all',
            doc=_(u'Retrieve and print all attributes from the server. Affects command output.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'raw',
            doc=_(u'Print entries as stored on the server. Only affects output format.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'pkey_only',
            required=False,
            label=_(u'Primary key only'),
            doc=_(u'Results should contain primary key attribute only ("anchor")'),
            default=False,
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.ListOfEntries(
            'result',
        ),
        output.Output(
            'count',
            int,
            doc=_(u'Number of entries returned'),
        ),
        output.Output(
            'truncated',
            bool,
            doc=_(u'True if not all results were returned'),
        ),
    )


@register()
class idoverrideuser_mod(Method):
    __doc__ = _("Modify an User ID override.")

    takes_args = (
        parameters.Str(
            'idviewcn',
            cli_name='idview',
            label=_(u'ID View Name'),
        ),
        parameters.Str(
            'ipaanchoruuid',
            cli_name='anchor',
            label=_(u'Anchor to override'),
        ),
    )
    takes_options = (
        parameters.Str(
            'description',
            required=False,
            cli_name='desc',
            label=_(u'Description'),
        ),
        parameters.Str(
            'uid',
            required=False,
            cli_name='login',
            label=_(u'User login'),
            no_convert=True,
        ),
        parameters.Int(
            'uidnumber',
            required=False,
            cli_name='uid',
            label=_(u'UID'),
            doc=_(u'User ID Number'),
        ),
        parameters.Str(
            'gecos',
            required=False,
            label=_(u'GECOS'),
        ),
        parameters.Int(
            'gidnumber',
            required=False,
            label=_(u'GID'),
            doc=_(u'Group ID Number'),
        ),
        parameters.Str(
            'homedirectory',
            required=False,
            cli_name='homedir',
            label=_(u'Home directory'),
        ),
        parameters.Str(
            'loginshell',
            required=False,
            cli_name='shell',
            label=_(u'Login shell'),
        ),
        parameters.Str(
            'ipaoriginaluid',
            required=False,
            exclude=('cli', 'webui'),
        ),
        parameters.Str(
            'ipasshpubkey',
            required=False,
            multivalue=True,
            cli_name='sshpubkey',
            label=_(u'SSH public key'),
            no_convert=True,
        ),
        parameters.Str(
            'setattr',
            required=False,
            multivalue=True,
            doc=_(u'Set an attribute to a name/value pair. Format is attr=value.\nFor multi-valued attributes, the command replaces the values already present.'),
            exclude=('webui',),
        ),
        parameters.Str(
            'addattr',
            required=False,
            multivalue=True,
            doc=_(u'Add an attribute/value pair. Format is attr=value. The attribute\nmust be part of the schema.'),
            exclude=('webui',),
        ),
        parameters.Str(
            'delattr',
            required=False,
            multivalue=True,
            doc=_(u'Delete an attribute/value pair. The option will be evaluated\nlast, after all sets and adds.'),
            exclude=('webui',),
        ),
        parameters.Flag(
            'rights',
            label=_(u'Rights'),
            doc=_(u'Display the access rights of this entry (requires --all). See ipa man page for details.'),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'all',
            doc=_(u'Retrieve and print all attributes from the server. Affects command output.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'raw',
            doc=_(u'Print entries as stored on the server. Only affects output format.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Str(
            'rename',
            required=False,
            label=_(u'Rename'),
            doc=_(u'Rename the User ID override object'),
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.Entry(
            'result',
        ),
        output.PrimaryKey(
            'value',
            doc=_(u"The primary_key value of the entry, e.g. 'jdoe' for a user"),
        ),
    )


@register()
class idoverrideuser_show(Method):
    __doc__ = _("Display information about an User ID override.")

    takes_args = (
        parameters.Str(
            'idviewcn',
            cli_name='idview',
            label=_(u'ID View Name'),
        ),
        parameters.Str(
            'ipaanchoruuid',
            cli_name='anchor',
            label=_(u'Anchor to override'),
        ),
    )
    takes_options = (
        parameters.Flag(
            'rights',
            label=_(u'Rights'),
            doc=_(u'Display the access rights of this entry (requires --all). See ipa man page for details.'),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'all',
            doc=_(u'Retrieve and print all attributes from the server. Affects command output.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'raw',
            doc=_(u'Print entries as stored on the server. Only affects output format.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.Entry(
            'result',
        ),
        output.PrimaryKey(
            'value',
            doc=_(u"The primary_key value of the entry, e.g. 'jdoe' for a user"),
        ),
    )


@register()
class idview_add(Method):
    __doc__ = _("Add a new ID View.")

    takes_args = (
        parameters.Str(
            'cn',
            cli_name='name',
            label=_(u'ID View Name'),
        ),
    )
    takes_options = (
        parameters.Str(
            'description',
            required=False,
            cli_name='desc',
            label=_(u'Description'),
        ),
        parameters.Str(
            'setattr',
            required=False,
            multivalue=True,
            doc=_(u'Set an attribute to a name/value pair. Format is attr=value.\nFor multi-valued attributes, the command replaces the values already present.'),
            exclude=('webui',),
        ),
        parameters.Str(
            'addattr',
            required=False,
            multivalue=True,
            doc=_(u'Add an attribute/value pair. Format is attr=value. The attribute\nmust be part of the schema.'),
            exclude=('webui',),
        ),
        parameters.Flag(
            'all',
            doc=_(u'Retrieve and print all attributes from the server. Affects command output.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'raw',
            doc=_(u'Print entries as stored on the server. Only affects output format.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.Entry(
            'result',
        ),
        output.PrimaryKey(
            'value',
            doc=_(u"The primary_key value of the entry, e.g. 'jdoe' for a user"),
        ),
    )


@register()
class idview_apply(Method):
    __doc__ = _("Applies ID View to specified hosts or current members of specified hostgroups. If any other ID View is applied to the host, it is overriden.")

    takes_args = (
        parameters.Str(
            'cn',
            cli_name='name',
            label=_(u'ID View Name'),
        ),
    )
    takes_options = (
        parameters.Str(
            'host',
            required=False,
            multivalue=True,
            cli_name='hosts',
            label=_(u'hosts'),
            doc=_(u'Hosts to apply the ID View to'),
        ),
        parameters.Str(
            'hostgroup',
            required=False,
            multivalue=True,
            cli_name='hostgroups',
            label=_(u'hostgroups'),
            doc=_(u'Hostgroups to whose hosts apply the ID View to. Please note that view is not applied automatically to any hosts added to the hostgroup after running the idview-apply command.'),
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.Output(
            'succeeded',
            dict,
            doc=_(u'Hosts that this ID View was applied to.'),
        ),
        output.Output(
            'failed',
            dict,
            doc=_(u'Hosts or hostgroups that this ID View could not be applied to.'),
        ),
        output.Output(
            'completed',
            int,
            doc=_(u'Number of hosts the ID View was applied to:'),
        ),
    )


@register()
class idview_del(Method):
    __doc__ = _("Delete an ID View.")

    takes_args = (
        parameters.Str(
            'cn',
            multivalue=True,
            cli_name='name',
            label=_(u'ID View Name'),
        ),
    )
    takes_options = (
        parameters.Flag(
            'continue',
            doc=_(u"Continuous mode: Don't stop on errors."),
            default=False,
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.Output(
            'result',
            dict,
            doc=_(u'List of deletions that failed'),
        ),
        output.ListOfPrimaryKeys(
            'value',
        ),
    )


@register()
class idview_find(Method):
    __doc__ = _("Search for an ID View.")

    takes_args = (
        parameters.Str(
            'criteria',
            required=False,
            doc=_(u'A string searched in all relevant object attributes'),
        ),
    )
    takes_options = (
        parameters.Str(
            'cn',
            required=False,
            cli_name='name',
            label=_(u'ID View Name'),
        ),
        parameters.Str(
            'description',
            required=False,
            cli_name='desc',
            label=_(u'Description'),
        ),
        parameters.Int(
            'timelimit',
            required=False,
            label=_(u'Time Limit'),
            doc=_(u'Time limit of search in seconds'),
        ),
        parameters.Int(
            'sizelimit',
            required=False,
            label=_(u'Size Limit'),
            doc=_(u'Maximum number of entries returned'),
        ),
        parameters.Flag(
            'all',
            doc=_(u'Retrieve and print all attributes from the server. Affects command output.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'raw',
            doc=_(u'Print entries as stored on the server. Only affects output format.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'pkey_only',
            required=False,
            label=_(u'Primary key only'),
            doc=_(u'Results should contain primary key attribute only ("name")'),
            default=False,
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.ListOfEntries(
            'result',
        ),
        output.Output(
            'count',
            int,
            doc=_(u'Number of entries returned'),
        ),
        output.Output(
            'truncated',
            bool,
            doc=_(u'True if not all results were returned'),
        ),
    )


@register()
class idview_mod(Method):
    __doc__ = _("Modify an ID View.")

    takes_args = (
        parameters.Str(
            'cn',
            cli_name='name',
            label=_(u'ID View Name'),
        ),
    )
    takes_options = (
        parameters.Str(
            'description',
            required=False,
            cli_name='desc',
            label=_(u'Description'),
        ),
        parameters.Str(
            'setattr',
            required=False,
            multivalue=True,
            doc=_(u'Set an attribute to a name/value pair. Format is attr=value.\nFor multi-valued attributes, the command replaces the values already present.'),
            exclude=('webui',),
        ),
        parameters.Str(
            'addattr',
            required=False,
            multivalue=True,
            doc=_(u'Add an attribute/value pair. Format is attr=value. The attribute\nmust be part of the schema.'),
            exclude=('webui',),
        ),
        parameters.Str(
            'delattr',
            required=False,
            multivalue=True,
            doc=_(u'Delete an attribute/value pair. The option will be evaluated\nlast, after all sets and adds.'),
            exclude=('webui',),
        ),
        parameters.Flag(
            'rights',
            label=_(u'Rights'),
            doc=_(u'Display the access rights of this entry (requires --all). See ipa man page for details.'),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'all',
            doc=_(u'Retrieve and print all attributes from the server. Affects command output.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'raw',
            doc=_(u'Print entries as stored on the server. Only affects output format.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Str(
            'rename',
            required=False,
            label=_(u'Rename'),
            doc=_(u'Rename the ID View object'),
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.Entry(
            'result',
        ),
        output.PrimaryKey(
            'value',
            doc=_(u"The primary_key value of the entry, e.g. 'jdoe' for a user"),
        ),
    )


@register()
class idview_show(Method):
    __doc__ = _("Display information about an ID View.")

    takes_args = (
        parameters.Str(
            'cn',
            cli_name='name',
            label=_(u'ID View Name'),
        ),
    )
    takes_options = (
        parameters.Flag(
            'rights',
            label=_(u'Rights'),
            doc=_(u'Display the access rights of this entry (requires --all). See ipa man page for details.'),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'show_hosts',
            required=False,
            doc=_(u'Enumerate all the hosts the view applies to.'),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'all',
            doc=_(u'Retrieve and print all attributes from the server. Affects command output.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'raw',
            doc=_(u'Print entries as stored on the server. Only affects output format.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.Entry(
            'result',
        ),
        output.PrimaryKey(
            'value',
            doc=_(u"The primary_key value of the entry, e.g. 'jdoe' for a user"),
        ),
    )


@register()
class idview_unapply(Method):
    __doc__ = _("Clears ID View from specified hosts or current members of specified hostgroups.")

    takes_options = (
        parameters.Str(
            'host',
            required=False,
            multivalue=True,
            cli_name='hosts',
            label=_(u'hosts'),
            doc=_(u'Hosts to clear (any) ID View from.'),
        ),
        parameters.Str(
            'hostgroup',
            required=False,
            multivalue=True,
            cli_name='hostgroups',
            label=_(u'hostgroups'),
            doc=_(u'Hostgroups whose hosts should have ID Views cleared. Note that view is not cleared automatically from any host added to the hostgroup after running idview-unapply command.'),
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.Output(
            'succeeded',
            dict,
            doc=_(u'Hosts that ID View was cleared from.'),
        ),
        output.Output(
            'failed',
            dict,
            doc=_(u'Hosts or hostgroups that ID View could not be cleared from.'),
        ),
        output.Output(
            'completed',
            int,
            doc=_(u'Number of hosts that had a ID View was unset:'),
        ),
    )



#
# Copyright (C) 2019  FreeIPA Contributors see COPYING for license
#
"""ctypes wrapper for libldap_str2dn
"""
from __future__ import absolute_import

import ctypes
import ctypes.util

import six

__all__ = ("str2dn", "dn2str", "DECODING_ERROR", "LDAPError")

# load reentrant libldap
ldap_r_lib = ctypes.util.find_library("ldap_r")
if ldap_r_lib is None:
    raise ImportError("libldap_r shared library missing")
try:
    lib = ctypes.CDLL(ldap_r_lib)
except OSError as e:
    raise ImportError(str(e))

# constants
LDAP_AVA_FREE_ATTR = 0x0010
LDAP_AVA_FREE_VALUE = 0x0020
LDAP_DECODING_ERROR = -4

# mask for AVA flags
AVA_MASK = ~(LDAP_AVA_FREE_ATTR | LDAP_AVA_FREE_VALUE)


class berval(ctypes.Structure):
    __slots__ = ()
    _fields_ = [("bv_len", ctypes.c_ulong), ("bv_value", ctypes.c_char_p)]

    def __bytes__(self):
        buf = ctypes.create_string_buffer(self.bv_value, self.bv_len)
        return buf.raw

    def __str__(self):
        return self.__bytes__().decode("utf-8")

    if six.PY2:
        __unicode__ = __str__
        __str__ = __bytes__


class LDAPAVA(ctypes.Structure):
    __slots__ = ()
    _fields_ = [
        ("la_attr", berval),
        ("la_value", berval),
        ("la_flags", ctypes.c_uint16),
    ]


# typedef LDAPAVA** LDAPRDN;
LDAPRDN = ctypes.POINTER(ctypes.POINTER(LDAPAVA))
# typedef LDAPRDN* LDAPDN;
LDAPDN = ctypes.POINTER(LDAPRDN)


def errcheck(result, func, arguments):
    if result != 0:
        if result == LDAP_DECODING_ERROR:
            raise DECODING_ERROR
        else:
            msg = ldap_err2string(result)
            raise LDAPError(msg.decode("utf-8"))
    return result


ldap_str2dn = lib.ldap_str2dn
ldap_str2dn.argtypes = (
    ctypes.c_char_p,
    ctypes.POINTER(LDAPDN),
    ctypes.c_uint16,
)
ldap_str2dn.restype = ctypes.c_int16
ldap_str2dn.errcheck = errcheck

ldap_dnfree = lib.ldap_dnfree
ldap_dnfree.argtypes = (LDAPDN,)
ldap_dnfree.restype = None

ldap_err2string = lib.ldap_err2string
ldap_err2string.argtypes = (ctypes.c_int16,)
ldap_err2string.restype = ctypes.c_char_p


class LDAPError(Exception):
    pass


class DECODING_ERROR(LDAPError):
    pass


# RFC 4514, 2.4
_ESCAPE_CHARS = {'"', "+", ",", ";", "<", ">", "'", "\x00"}


def _escape_dn(dn):
    if not dn:
        return ""
    result = []
    # a space or number sign occurring at the beginning of the string
    if dn[0] in {"#", " "}:
        result.append("\\")
    for c in dn:
        if c in _ESCAPE_CHARS:
            result.append("\\")
        result.append(c)
    # a space character occurring at the end of the string
    if len(dn) > 1 and result[-1] == " ":
        # insert before last entry
        result.insert(-1, "\\")
    return "".join(result)


def dn2str(dn):
    return ",".join(
        "+".join(
            "=".join((attr, _escape_dn(value))) for attr, value, _flag in rdn
        )
        for rdn in dn
    )


def str2dn(dn, flags=0):
    if dn is None:
        return []
    if isinstance(dn, six.text_type):
        dn = dn.encode("utf-8")

    ldapdn = LDAPDN()
    try:
        ldap_str2dn(dn, ctypes.byref(ldapdn), flags)

        result = []
        if not ldapdn:
            # empty DN, str2dn("") == []
            return result

        for rdn in ldapdn:
            if not rdn:
                break
            avas = []
            for ava_p in rdn:
                if not ava_p:
                    break
                ava = ava_p[0]
                avas.append(
                    (
                        six.text_type(ava.la_attr),
                        six.text_type(ava.la_value),
                        ava.la_flags & AVA_MASK,
                    )
                )
            result.append(avas)

        return result
    finally:
        ldap_dnfree(ldapdn)



# Authors:
#   Rob Crittenden <rcritten@redhat.com>
#
# Copyright (C) 2010  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Test the `ipalib.plugins.hbacsvcgroup` module.
"""

from ipalib import api, errors
from ipatests.test_xmlrpc.xmlrpc_test import Declarative, fuzzy_uuid
from ipatests.test_xmlrpc import objectclasses
from ipapython.dn import DN
import pytest

hbacsvcgroup1 = u'testhbacsvcgroup1'
dn1 = DN(('cn',hbacsvcgroup1),('cn','hbacservicegroups'),('cn','hbac'),
         api.env.basedn)

hbacsvc1 = u'sshd'
hbacsvc_dn1 = DN(('cn',hbacsvc1),('cn','hbacservices'),('cn','hbac'),
                 api.env.basedn)


@pytest.mark.tier1
class test_hbacsvcgroup(Declarative):

    cleanup_commands = [
        ('hbacsvcgroup_del', [hbacsvcgroup1], {}),
        ('hbacsvc_del', [hbacsvc1], {}),
    ]

    tests=[

        dict(
            desc='Try to retrieve non-existent %r' % hbacsvcgroup1,
            command=('hbacsvcgroup_show', [hbacsvcgroup1], {}),
            expected=errors.NotFound(
                reason=u'%s: HBAC service group not found' % hbacsvcgroup1),
        ),


        dict(
            desc='Try to update non-existent %r' % hbacsvcgroup1,
            command=('hbacsvcgroup_mod', [hbacsvcgroup1],
                dict(description=u'Updated hbacsvcgroup 1')
            ),
            expected=errors.NotFound(
                reason=u'%s: HBAC service group not found' % hbacsvcgroup1),
        ),


        dict(
            desc='Try to delete non-existent %r' % hbacsvcgroup1,
            command=('hbacsvcgroup_del', [hbacsvcgroup1], {}),
            expected=errors.NotFound(
                reason=u'%s: HBAC service group not found' % hbacsvcgroup1),
        ),


        dict(
            desc='Create %r' % hbacsvcgroup1,
            command=('hbacsvcgroup_add', [hbacsvcgroup1],
                dict(description=u'Test hbacsvcgroup 1')
            ),
            expected=dict(
                value=hbacsvcgroup1,
                summary=u'Added HBAC service group "testhbacsvcgroup1"',
                result=dict(
                    dn=dn1,
                    cn=[hbacsvcgroup1],
                    objectclass=objectclasses.hbacsvcgroup,
                    description=[u'Test hbacsvcgroup 1'],
                    ipauniqueid=[fuzzy_uuid],
                ),
            ),
        ),


        dict(
            desc='Try to create duplicate %r' % hbacsvcgroup1,
            command=('hbacsvcgroup_add', [hbacsvcgroup1],
                dict(description=u'Test hbacsvcgroup 1')
            ),
            expected=errors.DuplicateEntry(
                message=u'HBAC service group with name "%s" already exists' %
                    hbacsvcgroup1),
        ),


        dict(
            desc='Create service %r' % hbacsvc1,
            command=('hbacsvc_add', [hbacsvc1],
                dict(
                    description=u'Test service 1',
                ),
            ),
            expected=dict(
                value=hbacsvc1,
                summary=u'Added HBAC service "%s"' % hbacsvc1,
                result=dict(
                    dn=hbacsvc_dn1,
                    cn=[hbacsvc1],
                    description=[u'Test service 1'],
                    objectclass=objectclasses.hbacsvc,
                    ipauniqueid=[fuzzy_uuid],
                ),
            ),
        ),


        dict(
            desc=u'Add service %r to %r' % (hbacsvc1, hbacsvcgroup1),
            command=(
                'hbacsvcgroup_add_member', [hbacsvcgroup1], dict(hbacsvc=hbacsvc1)
            ),
            expected=dict(
                completed=1,
                failed=dict(
                    member=dict(
                        hbacsvc=tuple(),
                    ),
                ),
                result={
                    'dn': dn1,
                    'cn': [hbacsvcgroup1],
                    'description': [u'Test hbacsvcgroup 1'],
                    'member_hbacsvc': [hbacsvc1],
                },
            ),
        ),


        dict(
            desc='Retrieve %r' % hbacsvcgroup1,
            command=('hbacsvcgroup_show', [hbacsvcgroup1], {}),
            expected=dict(
                value=hbacsvcgroup1,
                summary=None,
                result={
                    'dn': dn1,
                    'member_hbacsvc': [hbacsvc1],
                    'cn': [hbacsvcgroup1],
                    'description': [u'Test hbacsvcgroup 1'],
                },
            ),
        ),


        dict(
            desc='Search for %r with members' % hbacsvcgroup1,
            command=('hbacsvcgroup_find', [], dict(
                cn=hbacsvcgroup1, no_members=False)),
            expected=dict(
                count=1,
                truncated=False,
                summary=u'1 HBAC service group matched',
                result=[
                    {
                        'dn': dn1,
                        'member_hbacsvc': [hbacsvc1],
                        'cn': [hbacsvcgroup1],
                        'description': [u'Test hbacsvcgroup 1'],
                    },
                ],
            ),
        ),


        dict(
            desc='Search for %r' % hbacsvcgroup1,
            command=('hbacsvcgroup_find', [], dict(cn=hbacsvcgroup1)),
            expected=dict(
                count=1,
                truncated=False,
                summary=u'1 HBAC service group matched',
                result=[
                    {
                        'dn': dn1,
                        'cn': [hbacsvcgroup1],
                        'description': [u'Test hbacsvcgroup 1'],
                    },
                ],
            ),
        ),


        dict(
            desc='Update %r' % hbacsvcgroup1,
            command=('hbacsvcgroup_mod', [hbacsvcgroup1],
                dict(description=u'Updated hbacsvcgroup 1')
            ),
            expected=dict(
                value=hbacsvcgroup1,
                summary=u'Modified HBAC service group "testhbacsvcgroup1"',
                result=dict(
                    cn=[hbacsvcgroup1],
                    description=[u'Updated hbacsvcgroup 1'],
                    member_hbacsvc=[hbacsvc1],
                ),
            ),
        ),


        dict(
            desc='Retrieve %r to verify update' % hbacsvcgroup1,
            command=('hbacsvcgroup_show', [hbacsvcgroup1], {}),
            expected=dict(
                value=hbacsvcgroup1,
                summary=None,
                result={
                    'dn': dn1,
                    'member_hbacsvc': [hbacsvc1],
                    'cn': [hbacsvcgroup1],
                    'description': [u'Updated hbacsvcgroup 1'],
                },
            ),
        ),


        dict(
            desc='Remove service %r from %r' % (hbacsvc1, hbacsvcgroup1),
            command=('hbacsvcgroup_remove_member', [hbacsvcgroup1],
                dict(hbacsvc=hbacsvc1)
            ),
            expected=dict(
                failed=dict(
                    member=dict(
                        hbacsvc=tuple(),
                    ),
                ),
                completed=1,
                result={
                    'dn': dn1,
                    'cn': [hbacsvcgroup1],
                    'description': [u'Updated hbacsvcgroup 1'],
                },
            ),
        ),


        dict(
            desc='Delete %r' % hbacsvcgroup1,
            command=('hbacsvcgroup_del', [hbacsvcgroup1], {}),
            expected=dict(
                value=[hbacsvcgroup1],
                summary=u'Deleted HBAC service group "testhbacsvcgroup1"',
                result=dict(failed=[]),
            ),
        ),


        dict(
            desc='Delete service %r' % hbacsvc1,
            command=('hbacsvc_del', [hbacsvc1], {}),
            expected=dict(
                value=[hbacsvc1],
                summary=u'Deleted HBAC service "%s"' % hbacsvc1,
                result=dict(failed=[]),
            ),
        )

    ]



#
# Copyright (C) 2018  FreeIPA Contributors see COPYING for license
#

ENTITY = 'config'

GRP_SEARCH_FIELD_DEFAULT = 'cn,description'
USR_SEARCH_FIELD_DEFAULT = 'uid,givenname,sn,telephonenumber,ou,title'

DATA = {
    'mod': [
        ('textbox', 'ipasearchrecordslimit', '200'),
        ('textbox', 'ipasearchtimelimit', '3'),
    ],
}

DATA2 = {
    'mod': [
        ('textbox', 'ipasearchrecordslimit', '100'),
        ('textbox', 'ipasearchtimelimit', '2'),
    ],
}

DATA_SIZE_LIMIT_LETTER = {
    'mod': [
        ('textbox', 'ipasearchrecordslimit', 'a'),
    ],
}

DATA_SIZE_LIMIT_SPACE = {
    'mod': [
        ('textbox', 'ipasearchrecordslimit', ' space'),
    ],
}

DATA_SIZE_LIMIT_NEG = {
    'mod': [
        ('textbox', 'ipasearchrecordslimit', '-2'),
    ],
}

DATA_TIME_LIMIT_LETTER = {
    'mod': [
        ('textbox', 'ipasearchtimelimit', 'a'),
    ],
}

DATA_TIME_LIMIT_SPACE = {
    'mod': [
        ('textbox', 'ipasearchtimelimit', ' space'),
    ],
}

DATA_TIME_LIMIT_NEG = {
    'mod': [
        ('textbox', 'ipasearchtimelimit', '-2'),
    ],
}



# Authors:
#   John Dennis <jdennis@redhat.com>
#
# Copyright (C) 2010  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import print_function

# WARNING: Do not import ipa modules, this is also used as a
# stand-alone script (invoked from po Makefile).
import optparse  # pylint: disable=deprecated-module
import sys
import gettext
import re
import os
import traceback
import polib
from collections import namedtuple

import six

'''
We test our translations by taking the original untranslated string
(e.g. msgid) and prepend a prefix character and then append a suffix
character. The test consists of asserting that the first character in the
translated string is the prefix, the last character in the translated string
is the suffix and the everything between the first and last character exactly
matches the original msgid.

We use unicode characters not in the ascii character set for the prefix and
suffix to enhance the test. To make reading the translated string easier the
prefix is the unicode right pointing arrow and the suffix left pointing arrow,
thus the translated string looks like the original string enclosed in
arrows. In ASCII art the string "foo" would render as:
-->foo<--
'''

#-------------------------------------------------------------------------------

verbose = False
print_traceback = False
pedantic = False
show_strings = True

# Unicode right pointing arrow
prefix = u'\u2192'               # utf-8 == '\xe2\x86\x92'
# Unicode left pointing arrow
suffix = u'\u2190'               # utf-8 == '\xe2\x86\x90'

page_width = 80
section_seperator = '=' * page_width
entry_seperator = '-' * page_width

# Python 3: Enforce ASCII mode so \w matches only ASCII chars. This avoids
# false positives in Chinese translation.
ASCII = getattr(re, "ASCII", 0)

# --------------------------------------------------------------------------
# For efficiency compile these regexps just once
_substitution_regexps = [
    re.compile(r'%[srduoxf]'),                 # e.g. %s
    re.compile(r'%\(\w+\)[srduoxf]', ASCII),   # e.g. %(foo)s
    re.compile(r'\$\w+', ASCII),               # e.g. $foo
    re.compile(r'\${\w+}', ASCII),             # e.g. ${foo}
    re.compile(r'\$\(\w+\)', ASCII)            # e.g. $(foo)
]
# Python style substitution, e.g. %(foo)s
# where foo is the key and s is the format char
# group 1: whitespace between % and (
# group 2: whitespace between ( and key
# group 3: whitespace between key and )
# group 4: whitespace between ) and format char
# group 5: format char
_python_substitution_regexp = re.compile(
    r'%(\s*)\((\s*)\w+(\s*)\)(\s*)([srduoxf])?', ASCII
)

# Shell style substitution, e.g. $foo $(foo) ${foo}
# where foo is the variable
_shell_substitution_regexp = re.compile(
    r'\$(\s*)([({]?)(\s*)\w+(\s*)([)}]?)', ASCII
)
# group 1: whitespace between $ and delimiter
# group 2: begining delimiter
# group 3: whitespace between beginning delmiter and variable
# group 4: whitespace between variable and ending delimiter
# group 5: ending delimiter

printf_fmt_re = re.compile(
    r"%"                                      # start
    r"(\d+\$)?"                               # fmt_arg    (group  1)
    r"(([#0 +'I]|-(?!\d))*)"                  # flags      (group  2)
    r"(([+-]?([1-9][0-9]*)?)|(\*|\*\d+\$))?"  # width      (group  4)
    r"(\.((-?\d*)|(\*|)|(\*\d+\$)))?"         # precision  (group  8)
    r"(h|hh|l|ll|L|j|z|t)?"                   # length     (group 13)
    r"([diouxXeEfFgGaAcspnm%])")              # conversion (group 14)

#-------------------------------------------------------------------------------

def get_prog_langs(entry):
    '''
    Given an entry in a pot or po file return a set of the
    programming languges it was found in. It needs to be a set
    because the same msgid may appear in more than one file which may
    be in different programming languages.

    Note: One might think you could use the c-format etc. flags to
    attached to entry to make this determination, but you can't. Those
    flags refer to the style of the string not the programming
    language it came from. Also the flags are often omitted and/or are
    inaccurate.

    For now we just look at the file extension. If we knew the path to
    the file we could use other heuristics such as looking for the
    shbang interpreter string.

    The set of possible language types witch might be returned are:

    * c
    * python

    '''
    result = set()

    for location in entry.occurrences:
        filename = location[0]
        ext = os.path.splitext(filename)[1]

        if ext in ('.c', '.h', '.cxx', '.cpp', '.hxx'):
            result.add('c')
        elif ext in ('.py'):
            result.add('python')

    return result

def parse_printf_fmt(s):
    '''
    Parse a printf style format string and return a list of format
    conversions found in the string.

    Each conversion specification is introduced by the character %, and
    ends with a conversion specifier.  In between there may be (in this
    order) zero or more flags, an optional minimum field width, an
    optional precision and an optional length modifier. See "man 3
    printf" for details.

    Each item in the returned list is a dict whose keys are the
    sub-parts of a conversion specification. The key and values are:

    fmt
        The entire format conversion specification
    fmt_arg
        The positional index of the matching argument in the argument
        list, e.g. %1$ indicates the first argument in the argument
        will be read for this conversion, excludes the leading % but
        includes the trailing $, 1$ is the fmt_arg in %1$.
    flags
        The flag characaters, e.g. 0 is the flag in %08d
    width
        The width field, e.g. 20 is the width in %20s
    precision
        The precisioin field, e.g. .2 is the precision in %8.2f
    length
        The length modifier field, e.g. l is the length modifier in %ld
    conversion
        The conversion specifier character, e.g. d is the conversion
        specification character in %ld

    If the part is not found in the format it's value will be None.
    '''

    result = []

    # get list of all matches, but skip escaped %
    matches = [x for x in printf_fmt_re.finditer(s) if x.group(0) != "%%"]

    # build dict of each sub-part of the format, append to result
    for match in matches:
        parts = {}
        parts['fmt']        = match.group(0)
        parts['fmt_arg']    = match.group(1)
        parts['flags']      = match.group(2) or None
        parts['width']      = match.group(4) or None
        parts['precision']  = match.group(8)
        parts['length']     = match.group(13)
        parts['conversion'] = match.group(14)

        result.append(parts)

    return result

def validate_substitutions_match(s1, s2, s1_name='string1', s2_name='string2'):
    '''
    Validate both s1 and s2 have the same number of substitution strings.
    A substitution string would be something that looked like this:

    * %(foo)s
    * $foo
    * ${foo}
    * $(foo)

    The substitutions may appear in any order in s1 and s2, however their
    format must match exactly and the exact same number of each must exist
    in both s1 and s2.

    A list of error diagnostics is returned explaining how s1 and s2 failed
    the validation check. If the returned error list is empty then the
    validation succeeded.

    :param s1:      First string to validate
    :param s2:      First string to validate
    :param s1_name: In diagnostic messages the name for s1
    :param s2_name: In diagnostic messages the name for s2
    :return:        List of diagnostic error messages, if empty then success
    '''
    errors = []

    def get_subs(s):
        '''
        Return a dict whoses keys are each unique substitution and whose
        value is the count of how many times that substitution appeared.
        '''
        subs = {}
        for regexp in _substitution_regexps:
            for match in regexp.finditer(s):
                matched = match.group(0)
                subs[matched] = subs.get(matched, 0) + 1
        return subs

    # Get the substitutions and their occurance counts
    subs1 = get_subs(s1)
    subs2 = get_subs(s2)

    # Form a set for each strings substitutions and
    # do set subtraction and interesection
    set1 = set(subs1.keys())
    set2 = set(subs2.keys())

    missing1 = set2 - set1
    missing2 = set1 - set2
    common = set1 & set2

    # Test for substitutions which are absent in either string
    if missing1:
        errors.append("The following substitutions are absent in %s: %s" %
                      (s1_name, ' '.join(missing1)))

    if missing2:
        errors.append("The following substitutions are absent in %s: %s" %
                      (s2_name, ' '.join(missing2)))

    if pedantic:
        # For the substitutions which are shared assure they occur an equal number of times
        for sub in common:
            if subs1[sub] != subs2[sub]:
                errors.append("unequal occurances of '%s', %s has %d occurances, %s has %d occurances" %
                              (sub, s1_name, subs1[sub], s2_name, subs2[sub]))

    if errors:
        if show_strings:
            errors.append('>>> %s <<<' % s1_name)
            errors.append(s1.rstrip())

            errors.append('>>> %s <<<' % s2_name)
            errors.append(s2.rstrip())
    return errors


def validate_substitution_syntax(s, s_name='string'):
    '''
    If s has one or more substitution variables then validate they
    are syntactically correct.
    A substitution string would be something that looked like this:

    * %(foo)s
    * $foo
    * ${foo}
    * $(foo)

    A list of error diagnostics is returned explaining how s1 and s2 failed
    the validation check. If the returned error list is empty then the
    validation succeeded.

    :param s:      String to validate
    :param s_name: In diagnostic messages the name for s
    :return:       List of diagnostic error messages, if empty then success
    '''
    errors = []

    # Look for Python style substitutions, e.g. %(foo)s
    for match in _python_substitution_regexp.finditer(s):
        if match.group(1):
            errors.append("%s has whitespace between %% and key in '%s'" %
                          (s_name, match.group(0)))
        if match.group(2) or match.group(3):
            errors.append("%s has whitespace next to key in '%s'" %
                          (s_name, match.group(0)))
        if match.group(4):
            errors.append("%s has whitespace between key and format character in '%s'" %
                          (s_name, match.group(0)))
        if not match.group(5):
            errors.append("%s has no format character in '%s'" %
                          (s_name, match.group(0)))

    # Look for shell style substitutions, e.g. $foo $(foo) ${foo}
    for match in _shell_substitution_regexp.finditer(s):
        if match.group(1):
            errors.append("%s has whitespace between $ and variable in '%s'" %
                          (s_name, match.group(0)))
        if match.group(3) or (match.group(4) and match.group(5)):
            errors.append("%s has whitespace next to variable in '%s'" %
                          (s_name, match.group(0)))

        beg_delimiter = match.group(2)
        end_delimiter = match.group(5)
        matched_delimiters = {'': '', '(': ')', '{': '}'}
        if beg_delimiter is not None or end_delimiter is not None:
            if matched_delimiters[beg_delimiter] != end_delimiter:
                errors.append("%s variable delimiters do not match in '%s', begin delimiter='%s' end delimiter='%s'" %
                              (s_name, match.group(0), beg_delimiter, end_delimiter))

    if errors:
        if show_strings:
            errors.append('>>> %s <<<' % s_name)
            errors.append(s.rstrip())

    return errors


def validate_positional_substitutions(s, prog_langs, s_name='string'):
    '''
    We do not permit multiple positional substitutions in translation
    strings (e.g. '%s') because they do not allow translators to reorder the
    wording. Instead keyword substitutions should be used when there are
    more than one.
    '''
    errors = []

    fmts = parse_printf_fmt(s)
    n_fmts = len(fmts)

    errors = []
    if n_fmts > 1:
        for fmt_parts in fmts:
            fmt        = fmt_parts['fmt']
            fmt_arg    = fmt_parts['fmt_arg']
            width      = fmt_parts['width']

            if width == '*':
                errors.append("Error: * width arg in format '%s should be indexed" % fmt)

            if fmt_arg is None:
                if 'c' in prog_langs:
                    errors.append("%s format '%s' is positional, should use indexed argument" %
                                  (s_name, fmt))
                else:
                    errors.append("%s format '%s' is positional, should use keyword substitution" %
                                  (s_name, fmt))

    if errors:
        if show_strings:
            errors.append('>>> %s <<<' % s_name)
            errors.append(s.rstrip())

    return errors

def validate_file(file_path, validation_mode, reference_pot=None):
    '''
    Given a pot or po file scan all it's entries looking for problems
    with variable substitutions. See the following functions for
    details on how the validation is performed.

    * validate_substitutions_match()
    * validate_substitution_syntax()
    * validate_positional_substitutions()

    Returns the number of entries with errors.

    For po files, ``reference_pot`` gives a pot file to merge with (to recover
    comments and file locations)
    '''

    def emit_messages():
        if n_warnings:
            warning_lines.insert(0, section_seperator)
            warning_lines.insert(1, "%d validation warnings in %s" % (n_warnings, file_path))
            print('\n'.join(warning_lines))

        if n_errors:
            error_lines.insert(0, section_seperator)
            error_lines.insert(1, "%d validation errors in %s" % (n_errors, file_path))
            print('\n'.join(error_lines))

    Result = namedtuple('ValidateFileResult', ['n_entries', 'n_msgids', 'n_msgstrs', 'n_warnings', 'n_errors'])

    warning_lines = []
    error_lines = []
    n_entries = 0
    n_msgids = 0
    n_msgstrs = 0
    n_entries = 0
    n_warnings = 0
    n_errors  = 0
    n_plural_forms = 0

    if not os.path.isfile(file_path):
        error_lines.append(entry_seperator)
        error_lines.append('file does not exist "%s"' % (file_path))
        n_errors += 1
        emit_messages()
        return Result(n_entries=n_entries, n_msgids=n_msgids, n_msgstrs=n_msgstrs, n_warnings=n_warnings, n_errors=n_errors)

    try:
        po = polib.pofile(file_path)
    except Exception as e:
        error_lines.append(entry_seperator)
        error_lines.append('Unable to parse file "%s": %s' % (file_path, e))
        n_errors += 1
        emit_messages()
        return Result(n_entries=n_entries, n_msgids=n_msgids, n_msgstrs=n_msgstrs, n_warnings=n_warnings, n_errors=n_errors)

    if validation_mode == 'po' and reference_pot:
        # Merge the .pot file for comments and file locations
        po.merge(reference_pot)

    if validation_mode == 'po':
        plural_forms = po.metadata.get('Plural-Forms')
        if not plural_forms:
            error_lines.append(entry_seperator)
            error_lines.append("%s: does not have Plural-Forms header" % file_path)
            n_errors += 1
        match = re.search(r'\bnplurals\s*=\s*(\d+)', plural_forms)
        if match:
            n_plural_forms = int(match.group(1))
        else:
            error_lines.append(entry_seperator)
            error_lines.append("%s: does not specify integer nplurals in Plural-Forms header" % file_path)
            n_errors += 1

    n_entries = len(po)
    for entry in po:
        entry_warnings = []
        entry_errors = []
        have_msgid = entry.msgid.strip() != ''
        have_msgid_plural = entry.msgid_plural.strip() != ''
        have_msgstr = entry.msgstr.strip() != ''

        if have_msgid:
            n_msgids += 1
        if have_msgid_plural:
            n_msgids += 1
        if have_msgstr:
            n_msgstrs += 1

        if validation_mode == 'pot':
            prog_langs = get_prog_langs(entry)
            if have_msgid:
                errors = validate_positional_substitutions(entry.msgid, prog_langs, 'msgid')
                entry_errors.extend(errors)
            if have_msgid_plural:
                errors = validate_positional_substitutions(entry.msgid_plural, prog_langs, 'msgid_plural')
                entry_errors.extend(errors)
        elif validation_mode == 'po':
            if have_msgid:
                if have_msgstr:
                    errors = validate_substitutions_match(entry.msgid, entry.msgstr, 'msgid', 'msgstr')
                    entry_errors.extend(errors)

                if have_msgid_plural and have_msgstr:
                    n_plurals = 0
                    for index, msgstr in entry.msgstr_plural.items():
                        have_msgstr_plural = msgstr.strip() != ''
                        if have_msgstr_plural:
                            n_plurals += 1
                            errors = validate_substitutions_match(entry.msgid_plural, msgstr, 'msgid_plural', 'msgstr_plural[%s]' % index)
                            entry_errors.extend(errors)
                        else:
                            entry_errors.append('msgstr_plural[%s] is empty' % (index))
                    if n_plural_forms != n_plurals:
                        entry_errors.append('%d plural forms specified, but this entry has %d plurals' % (n_plural_forms, n_plurals))

        if pedantic:
            if have_msgid:
                errors = validate_substitution_syntax(entry.msgid, 'msgid')
                entry_warnings.extend(errors)

            if have_msgid_plural:
                errors = validate_substitution_syntax(entry.msgid_plural, 'msgid_plural')
                entry_warnings.extend(errors)

                errors = validate_substitutions_match(entry.msgid, entry.msgid_plural, 'msgid', 'msgid_plural')
                entry_warnings.extend(errors)

                for index, msgstr in entry.msgstr_plural.items():
                    have_msgstr_plural = msgstr.strip() != ''
                    if have_msgstr_plural:
                        errors = validate_substitution_syntax(msgstr,  'msgstr_plural[%s]' % index)
                        entry_warnings.extend(errors)

            if have_msgstr:
                errors = validate_substitution_syntax(entry.msgstr, 'msgstr')
                entry_warnings.extend(errors)

        if entry_warnings:
            warning_lines.append(entry_seperator)
            warning_lines.append('locations: %s' % (', '.join(["%s:%d" % (x[0], int(x[1])) for x in entry.occurrences])))
            warning_lines.extend(entry_warnings)
            n_warnings += 1

        if entry_errors:
            error_lines.append(entry_seperator)
            error_lines.append('locations: %s' % (', '.join(["%s:%d" % (x[0], int(x[1])) for x in entry.occurrences])))
            error_lines.extend(entry_errors)
            n_errors += 1

    emit_messages()
    return Result(n_entries=n_entries, n_msgids=n_msgids, n_msgstrs=n_msgstrs, n_warnings=n_warnings, n_errors=n_errors)


#----------------------------------------------------------------------
def create_po(pot_file, po_file, mo_file):

    if not os.path.isfile(pot_file):
        print('file does not exist "%s"' % (pot_file), file=sys.stderr)
        return 1
    try:
        po = polib.pofile(pot_file)
    except Exception as e:
        print('Unable to parse file "%s": %s' % (pot_file, e), file=sys.stderr)
        return 1

    # Update the metadata in the po file header
    # It's case insensitive so search the keys in a case insensitive manner
    #
    # We need to update the Plural-Forms otherwise gettext.py will raise the
    # following error:
    #
    # raise ValueError, 'plural forms expression could be dangerous'
    #
    # It is demanding the rhs of plural= only contains the identifer 'n'

    for k in po.metadata:
        if k.lower() == 'plural-forms':
            po.metadata[k] = 'nplurals=2; plural=(n != 1)'
        # the auto-generated PO file should have charset set to UTF-8
        # because we are using UTF-8 prefix and suffix below
        elif k.lower() == 'content-type':
            po.metadata[k] = 'Content-Type: text/plain; charset=UTF-8'


    # Iterate over all msgid's and form a msgstr by prepending
    # the prefix and appending the suffix
    for entry in po:
        if entry.msgid_plural:
            entry.msgstr_plural = {0: prefix + entry.msgid + suffix,
                                   1: prefix + entry.msgid_plural + suffix}
        else:
            entry.msgstr = prefix + entry.msgid + suffix

    # Write out the po and mo files
    po.save(po_file)
    print("Wrote: %s" % (po_file))

    po.save_as_mofile(mo_file)
    print("Wrote: %s" % (mo_file))

    return 0

#----------------------------------------------------------------------

def validate_unicode_edit(msgid, msgstr):
    # Verify the first character is the test prefix
    if msgstr[0] != prefix:
        raise ValueError('First char in translated string "%s" not equal to prefix "%s"' %
                         (msgstr.encode('utf-8'), prefix.encode('utf-8')))

    # Verify the last character is the test suffix
    if msgstr[-1] != suffix:
        raise ValueError('Last char in translated string "%s" not equal to suffix "%s"' %
                         (msgstr.encode('utf-8'), suffix.encode('utf-8')))

    # Verify everything between the first and last character is the
    # original untranslated string
    if msgstr[1:-1] != msgid:
        raise ValueError('Translated string "%s" minus the first & last character is not equal to msgid "%s"' %
                         (msgstr.encode('utf-8'), msgid))

    if verbose:
        msg = 'Success: message string "%s" maps to translated string "%s"' % (msgid, msgstr)
        print(msg.encode('utf-8'))


def test_translations(po_file, lang, domain, locale_dir):
    # The test installs the test message catalog under the xh_ZA
    # (e.g. Zambia Xhosa) language by default. It would be nice to
    # use a dummy language not associated with any real language,
    # but the setlocale function demands the locale be a valid
    # known locale, Zambia Xhosa is a reasonable choice :)
    locale_envs = ('LANGUAGE', 'LC_ALL', 'LC_MESSAGES', 'LANG')

    os.environ.update(
        {locale_env: lang for locale_env in locale_envs}
    )

    # Create a gettext translation object specifying our domain as
    # 'ipa' and the locale_dir as 'test_locale' (i.e. where to
    # look for the message catalog). Then use that translation
    # object to obtain the translation functions.

    t = gettext.translation(domain, locale_dir)

    if six.PY2:
        # pylint: disable=no-member
        get_msgstr = t.ugettext
        get_msgstr_plural = t.ungettext
        # pylint: enable=no-member
    else:
        get_msgstr = t.gettext
        get_msgstr_plural = t.ngettext

    return po_file_iterate(po_file, get_msgstr, get_msgstr_plural)

def po_file_iterate(po_file, get_msgstr, get_msgstr_plural):
    try:
        # Iterate over the msgid's
        if not os.path.isfile(po_file):
            print('file does not exist "%s"' % (po_file), file=sys.stderr)
            return 1
        try:
            po = polib.pofile(po_file)
        except Exception as e:
            print('Unable to parse file "%s": %s' % (po_file, e), file=sys.stderr)
            return 1

        n_entries = 0
        n_translations = 0
        n_valid = 0
        n_fail = 0
        for entry in po:
            if entry.msgid_plural:
                msgid = entry.msgid
                msgid_plural = entry.msgid_plural
                msgstr = get_msgstr_plural(msgid, msgid_plural, 1)
                msgstr_plural = get_msgstr_plural(msgid, msgid_plural, 2)

                try:
                    n_translations += 1
                    validate_unicode_edit(msgid, msgstr)
                    n_valid += 1
                except Exception as e:
                    n_fail += 1
                    if print_traceback:
                        traceback.print_exc()
                    print("ERROR: %s" % e, file=sys.stderr)

                try:
                    n_translations += 1
                    validate_unicode_edit(msgid_plural, msgstr_plural)
                    n_valid += 1
                except Exception as e:
                    n_fail += 1
                    if print_traceback:
                        traceback.print_exc()
                    print("ERROR: %s" % e, file=sys.stderr)


            else:
                msgid = entry.msgid
                msgstr = get_msgstr(msgid)

                try:
                    n_translations += 1
                    validate_unicode_edit(msgid, msgstr)
                    n_valid += 1
                except Exception as e:
                    n_fail += 1
                    if print_traceback:
                        traceback.print_exc()
                    print("ERROR: %s" % e, file=sys.stderr)

            n_entries += 1

    except Exception as e:
        if print_traceback:
            traceback.print_exc()
        print("ERROR: %s" % e, file=sys.stderr)
        return 1

    if not n_entries:
        print("ERROR: no translations found in %s" % (po_file), file=sys.stderr)
        return 1

    if n_fail:
        print("ERROR: %d failures out of %d translations" % (n_fail, n_entries), file=sys.stderr)
        return 1

    print("%d translations in %d messages successfully tested" % (n_translations, n_entries))
    return 0

#----------------------------------------------------------------------

usage ='''

%prog --test-gettext
%prog --create-test
%prog --validate-pot [pot_file1, ...]
%prog --validate-po po_file1 [po_file2, ...]
'''

def main():
    global verbose, print_traceback, pedantic, show_strings

    parser = optparse.OptionParser(usage=usage)

    mode_group = optparse.OptionGroup(parser, 'Operational Mode',
                                      'You must select one these modes to run in')

    mode_group.add_option('-g', '--test-gettext', action='store_const', const='test_gettext', dest='mode',
                          help='create the test translation file(s) and exercise them')
    mode_group.add_option('-c', '--create-test', action='store_const', const='create_test', dest='mode',
                          help='create the test translation file(s)')
    mode_group.add_option('-P', '--validate-pot', action='store_const', const='validate_pot', dest='mode',
                          help='validate pot file(s)')
    mode_group.add_option('-p', '--validate-po', action='store_const', const='validate_po', dest='mode',
                          help='validate po file(s)')

    parser.add_option_group(mode_group)
    parser.set_defaults(mode='')

    parser.add_option('-s', '--show-strings', action='store_true', dest='show_strings', default=False,
                      help='show the offending string when an error is detected')
    parser.add_option('--pedantic', action='store_true', dest='pedantic', default=False,
                      help='be aggressive when validating')
    parser.add_option('-v', '--verbose', action='store_true', dest='verbose', default=False,
                      help='be informative')
    parser.add_option('--traceback', action='store_true', dest='print_traceback', default=False,
                      help='print the traceback when an exception occurs')

    param_group = optparse.OptionGroup(parser, 'Run Time Parameters',
                                       'These may be used to modify the run time defaults')

    param_group.add_option('--test-lang', action='store', dest='test_lang', default='test',
                           help="test po file uses this as it's basename (default=test)")
    param_group.add_option('--lang', action='store', dest='lang', default='xh_ZA',
                           help='lang used for locale, MUST be a valid lang (default=xh_ZA)')
    param_group.add_option('--domain', action='store', dest='domain', default='ipa',
                           help='translation domain used during test (default=ipa)')
    param_group.add_option('--locale-dir', action='store', dest='locale_dir', default='test_locale',
                           help='locale directory used during test (default=test_locale)')
    param_group.add_option('--pot-file', action='store', dest='pot_file', default='ipa.pot',
                           help='default pot file, used when validating pot file or generating test po and mo files (default=ipa.pot)')

    parser.add_option_group(param_group)

    options, args = parser.parse_args()

    verbose = options.verbose
    print_traceback = options.print_traceback
    pedantic = options.pedantic
    show_strings = options.show_strings

    if not options.mode:
        print('ERROR: no mode specified', file=sys.stderr)
        return 1

    if options.mode in ('validate_pot', 'validate_po'):
        if options.mode == 'validate_pot':
            files = args
            if not files:
                files = [options.pot_file]
            validation_mode = 'pot'
            reference_pot = None
        elif options.mode == 'validate_po':
            files = args
            if not files:
                print('ERROR: no po files specified', file=sys.stderr)
                return 1
            validation_mode = 'po'
            reference_pot = polib.pofile(options.pot_file)
        else:
            print('ERROR: unknown validation mode "%s"' % (options.mode), file=sys.stderr)
            return 1

        total_entries = 0
        total_msgids = 0
        total_msgstrs = 0
        total_warnings = 0
        total_errors = 0

        for f in files:
            result = validate_file(f, validation_mode, reference_pot)
            total_entries += result.n_entries
            total_msgids += result.n_msgids
            total_msgstrs += result.n_msgstrs
            total_warnings += result.n_warnings
            total_errors += result.n_errors
            print("%s: %d entries, %d msgid, %d msgstr, %d warnings %d errors" % \
                (f, result.n_entries, result.n_msgids, result.n_msgstrs, result.n_warnings, result.n_errors))
        if total_errors:
            print(section_seperator)
            print("%d errors in %d files" % (total_errors, len(files)))
            return 1
        else:
            return 0

    elif options.mode in ('create_test', 'test_gettext'):
        po_file = '%s.po' % options.test_lang
        pot_file = options.pot_file

        msg_dir = os.path.join(options.locale_dir, options.lang, 'LC_MESSAGES')
        if not os.path.exists(msg_dir):
            os.makedirs(msg_dir)

        mo_basename = '%s.mo' % options.domain
        mo_file = os.path.join(msg_dir, mo_basename)

        result = create_po(pot_file, po_file, mo_file)
        if result:
            return result

        if options.mode == 'create_test':
            return result

        # The test installs the test message catalog under the xh_ZA
        # (e.g. Zambia Xhosa) language by default. It would be nice to
        # use a dummy language not associated with any real language,
        # but the setlocale function demands the locale be a valid
        # known locale, Zambia Xhosa is a reasonable choice :)

        lang = options.lang

        # Create a gettext translation object specifying our domain as
        # 'ipa' and the locale_dir as 'test_locale' (i.e. where to
        # look for the message catalog). Then use that translation
        # object to obtain the translation functions.

        domain = options.domain
        locale_dir = options.locale_dir

        return test_translations(po_file, lang, domain, locale_dir)

    else:
        print('ERROR: unknown mode "%s"' % (options.mode), file=sys.stderr)
        return 1

if __name__ == "__main__":
    sys.exit(main())



#
# Copyright (C) 2016  FreeIPA Contributors see COPYING for license
#

import importlib
import os
import re
import sys

import six

from ipaclient.frontend import ClientCommand, ClientMethod
from ipalib.frontend import Object
from ipapython.ipautil import APIVersion

if six.PY3:
    unicode = str


class CompatCommand(ClientCommand):
    @property
    def forwarded_name(self):
        return self.name


class CompatMethod(ClientMethod, CompatCommand):
    pass


class CompatObject(Object):
    pass


def get_package(server_info, client):
    try:
        server_version = server_info['version']
    except KeyError:
        is_valid = False
    else:
        is_valid = server_info.is_valid()

    if not is_valid:
        if not client.isconnected():
            client.connect(verbose=False)
        env = client.forward(u'env', u'api_version', version=u'2.0')
        try:
            server_version = env['result']['api_version']
        except KeyError:
            ping = client.forward(u'ping', version=u'2.0')
            try:
                match = re.search(u'API version (2\.[0-9]+)', ping['summary'])
            except KeyError:
                match = None
            if match is not None:
                server_version = match.group(1)
            else:
                server_version = u'2.0'
        server_info['version'] = server_version
        server_info.update_validity()

    server_version = APIVersion(server_version)

    package_names = {}
    base_name = __name__.rpartition('.')[0]
    base_dir = os.path.dirname(__file__)
    for name in os.listdir(base_dir):
        package_dir = os.path.join(base_dir, name)
        if name.startswith('2_') and os.path.isdir(package_dir):
            package_version = APIVersion(name.replace('_', '.'))
            package_names[package_version] = '{}.{}'.format(base_name, name)

    package_version = None
    for version in sorted(package_names):
        if package_version is None or package_version < version:
            package_version = version
        if version >= server_version:
            break

    package_name = package_names[package_version]
    try:
        package = sys.modules[package_name]
    except KeyError:
        package = importlib.import_module(package_name)

    return package



#
# Copyright (C) 2017  FreeIPA Contributors see COPYING for license
#

import six
from ipalib import api, Command, errors, output, Str
from ipalib import _
from ipapython.dn import DN
from ipalib.plugable import Registry
from .idviews import DEFAULT_TRUST_VIEW_NAME

if six.PY3:
    unicode = str

__doc__ = _("""
Return information about currently authenticated identity

Who am I command returns information on how to get
more details about the identity authenticated for this
request. The information includes:

 * type of object
 * command to retrieve details of the object
 * arguments and options to pass to the command

The information is returned as a dictionary. Examples below use
'key: value' output for illustrative purposes.

EXAMPLES:

 Look up as IPA user:
   kinit admin
   ipa console
   >> api.Command.whoami()
   ------------------------------------------
   object: user
   command: user_show/1
   arguments: admin
   ------------------------------------------

 Look up as a user from a trusted domain:
   kinit user@AD.DOMAIN
   ipa console
   >> api.Command.whoami()
   ------------------------------------------
   object: idoverrideuser
   command: idoverrideuser_show/1
   arguments: ('default trust view', 'user@ad.domain')
   ------------------------------------------

 Look up as a host:
   kinit -k
   ipa console
   >> api.Command.whoami()
   ------------------------------------------
   object: host
   command: host_show/1
   arguments: ipa.example.com
   ------------------------------------------

 Look up as a Kerberos service:
   kinit -k -t /path/to/keytab HTTP/ipa.example.com
   ipa console
   >> api.Command.whoami()
   ------------------------------------------
   object: service
   command: service_show/1
   arguments: HTTP/ipa.example.com
   ------------------------------------------
""")

register = Registry()


@register()
class whoami(Command):
    __doc__ = _('Describe currently authenticated identity.')

    NO_CLI = True

    output_params = (
        Str('object', label=_('Object class name')),
        Str('command', label= _('Function to get details')),
        Str('arguments*', label=_('Arguments to details function')),
    )

    has_output = (
        output.Output('object', unicode, _('Object class name')),
        output.Output('command', unicode, _('Function to get details')),
        output.Output('arguments', (list, tuple),
                      _('Arguments to details function')),
    )

    def execute(self, **options):
        """
        Retrieve the DN we are authenticated as to LDAP and find bindable IPA
        object that handles the container where this DN belongs to. Then report
        details about this object.
        """
        exceptions = {
                'idoverrideuser': (DN("cn={0}".format(DEFAULT_TRUST_VIEW_NAME)),
                                   DEFAULT_TRUST_VIEW_NAME, 'ipaOriginalUid'),
        }
        ldap = api.Backend.ldap2

        # whoami_s() call returns a string 'dn: <actual DN value>'
        # We also reject ldapi-as-root connections as DM is a virtual object
        dn = DN(ldap.conn.whoami_s()[4:])
        if dn == DN('cn=Directory Manager'):
            raise errors.NotFound(
                    reason=_('Cannot query Directory Manager with API'))

        entry = ldap.get_entry(dn)
        o_name = None
        o_func = None
        o_args = []
        for o in api.Object():
            if not getattr(o, 'bindable', None):
                continue
            container = getattr(o, 'container_dn', None)
            if container is None:
                continue
            # Adjust container for exception two-level objects
            if o.name in exceptions:
                container = exceptions[o.name][0] + container
            if dn.find(container + api.env.basedn) == 1:
                # We found exact container this DN belongs to
                o_name = unicode(o.name)
                o_args = [unicode(entry.single_value.get(o.primary_key.name))]
                o_func = unicode(o.methods.show.full_name)
                if o.name in exceptions:
                    o_args = [unicode(exceptions[o.name][1]),
                              unicode(entry.single_value.get(
                                      exceptions[o.name][2]))]
                break

        return {'object': o_name, 'command': o_func, 'arguments': o_args}
